import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


def main():
    os.environ["DJANGO_SETTINGS_MODULE"] = "wechat_django.tests.settings"
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(["wechat_django", "wechat_django.pay"])
    sys.exit(bool(failures))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from wechatpy import parse_message

from . import WeChatApp, WeChatUser


class WeChatInfo(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def app(self):
        """
        :rtype: wechat_django.models.WeChatApp
        """
        if not hasattr(self, "_app"):
            self._app = WeChatApp.objects.get_by_name(self.appname)
        return self._app

    @property
    def user(self):
        """
        :rtype: wechat_django.models.WeChatUser
        """
        if not hasattr(self, "_user"):
            raise NotImplementedError()
        return self._user

    @property
    def request(self):
        """
        :rtype: django.http.request.HttpRequest
        """
        return self._request

    @property
    def appname(self):
        return self._appname


class WeChatMessageInfo(WeChatInfo):
    """由微信接收到的消息"""
    @property
    def message(self):
        """
        :raises: xmltodict.expat.ExpatError
        :rtype: wechatpy.messages.BaseMessage
        """
        if not hasattr(self, "_message"):
            app = self.app
            request = self.request
            raw = self.raw
            if app.crypto:
                self._raw = app.crypto.decrypt_message(
                    self.raw,
                    request.GET["msg_signature"],
                    request.GET["timestamp"],
                    request.GET["nonce"]
                )
            self._message = parse_message(self.raw)
        return self._message

    @property
    def local_user(self):
        """
        不从微信服务器重新同步用户
        :rtype: wechat_django.models.WeChatUser
        """
        if not hasattr(self, "_user") and not hasattr(self, "_local_user"):
            self._local_user = self.app.user_by_openid(
                self.message.source, ignore_errors=True, sync_user=False)
        return self._user if hasattr(self, "_user") else self._local_user

    @property
    def user(self):
        """
        :rtype: wechat_django.models.WeChatUser
        """
        if not hasattr(self, "_user"):
            self._user = self.app.user_by_openid(
                self.message.source, ignore_errors=True)
        return self._user

    @property
    def raw(self):
        """原始消息
        :rtype: str
        """
        if hasattr(self, "_raw"):
            return self._raw
        return self.request.body


class WeChatOAuthInfo(WeChatInfo):
    """附带在request上的微信对象
    """
    @property
    def scope(self):
        """授权的scope
        :rtype: tuple
        """
        from ..oauth import WeChatSNSScope

        if not hasattr(self, "_scope"):
            self._scope = (WeChatSNSScope.BASE,)
        return self._scope

    _state = ""
    @property
    def state(self):
        """授权携带的state"""
        return self._state

    @property
    def oauth_uri(self):
        return self.app.oauth.authorize_url(
            self.redirect_uri,
            ",".join(self.scope),
            self.state
        )

    @property
    def redirect_uri(self):
        """授权后重定向回的地址"""
        return self._redirect_uri

    @property
    def openid(self):
        if not hasattr(self, "_openid"):
            self._openid = self.request.get_signed_cookie(
                self.session_key, None)
        return self._openid

    @property
    def user(self):
        if not hasattr(self, "_user"):
            self._user = self.app.user_by_openid(self.openid)
        return super(WeChatOAuthInfo, self).user

    @property
    def session_key(self):
        return "wechat_{0}_user".format(self.appname)

    def __str__(self):
        return "WeChatOuathInfo: " + "\t".join(
            "{k}: {v}".format(k=attr, v=getattr(self, attr, None))
            for attr in
            ("app", "user", "redirect", "oauth_uri", "state", "scope")
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from six.moves import reduce
from six.moves.urllib.parse import parse_qsl


def anchor(text, href, **kwargs):
    """转化为a标签"""
    @mark_safe
    def wrapper(modeladmin, obj):
        kwargs.update(
            href=href,
            text=text
        )
        for key, value in kwargs.items():
            if callable(value):
                kwargs[key] = value(modeladmin, obj)
        return kwargs["text"] and '<a href="{href}">{text}</a>'.format(**kwargs)
    return wrapper


def foreignkey(field_name):
    """
    Converts a foreign key value into clickable links.

    If field_name is 'parent', link text will be str(obj.parent)
    Link will be admin url for the admin url for obj.parent.id:change
    """
    @mark_safe
    def _linkify(obj):
        app_label = obj._meta.app_label
        linked_obj = getattr(obj, field_name)
        model_name = linked_obj._meta.model_name
        view_name = "admin:{app_label}_{model_name}_change".format(
            app_label=app_label,
            model_name=model_name
        )
        link_url = reverse(view_name, kwargs=dict(
            object_id=linked_obj.id,
            wechat_app_id=obj.app_id
        ))
        return '<a href="{0}">{1}</a>'.format(link_url, linked_obj)

    _linkify.short_description = _(field_name)
    _linkify.admin_order_field = field_name
    return _linkify


def list_property(field_name, **kwargs):
    def _from_property(obj):
        rv = reduce(getattr, field_name.split("."), obj)
        return rv() if callable(rv) else rv

    for key, value in kwargs.items():
        setattr(_from_property, key, value)
    return _from_property


def field_property(field_name, **kwargs):
    def _from_property(admin, obj=None):
        if not obj:
            return None
        rv = reduce(getattr, field_name.split("."), obj)
        return rv() if callable(rv) else rv

    for key, value in kwargs.items():
        setattr(_from_property, key, value)
    return _from_property


def get_request_params(request, param):
    """从请求信息中获取想要的信息"""
    if not hasattr(request, param):
        preserved_filters_str = request.GET.get('_changelist_filters')
        if preserved_filters_str:
            preserved_filters = dict(parse_qsl(preserved_filters_str))
        else:
            preserved_filters = dict()
        value = (request.GET.get(param)
            or preserved_filters.get(param))
        setattr(request, param, value)
    return getattr(request, param)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from functools import update_wrapper, wraps

from django.db import models as m
from django.utils.encoding import force_bytes
import six

from . import WeChatApp


def create_shortcut(model):
    def method(func_or_name):
        def decorator(func):
            @wraps(func)
            def decorated_func(*args, **kwargs):
                return func(*args, **kwargs)
            decorated_func._shortcutmethod = name
            decorated_func._model = model

            return decorated_func

        if callable(func_or_name):
            name = func_or_name.__name__
            return decorator(func_or_name)
        else:
            name = func_or_name
            return decorator

    return method


appmethod = create_shortcut(WeChatApp)
"""可以把方法注册到WeChatApp上的语法糖

    class SomeModel(WeChatModel):
        @classmethod
        @appmethod
        def awesome_method(cls, app, *args, **kwargs):
            # ...

        @classmethod
        @appmethod("another_method")
        def method(cls, app, *args, **kwargs):
            # ...

    res = SomeModel.awesome_method(app, *args, **kwargs)
    res = app.awesome_method(*args, **kwargs)
    res = app.another_method(*args, **kwargs)
"""


class WeChatModelMetaClass(m.base.ModelBase):
    def __new__(meta, name, bases, attrs):
        # python2.7 __str__必须返回bytestring
        if six.PY2 and "__str__" in attrs:
            __str__ = attrs.pop("__str__")
            attrs["__str__"] = update_wrapper(
                lambda self: force_bytes(__str__(self)), __str__)

        cls = super(WeChatModelMetaClass, meta).__new__(
            meta, name, bases, attrs)
        return cls

    def __init__(cls, name, bases, attrs):
        super(WeChatModelMetaClass, cls).__init__(name, bases, attrs)

        def shortcutmethod(method):
            def wrapped_func(self, *args, **kwargs):
                return method(self, *args, **kwargs)

            method_name = method._shortcutmethod
            rv = wraps(method)(wrapped_func)
            rv.__name__ = str(method_name)
            rv.__qualname__ = "{0}.{1}".format(
                method._model.__name__, method_name)
            return rv

        for attr in attrs:
            # 将modelmethod转换为shortcut method
            val = getattr(cls, attr)
            if getattr(val, "_shortcutmethod", False):
                setattr(
                    val._model, val._shortcutmethod, shortcutmethod(val))


class WeChatModel(six.with_metaclass(WeChatModelMetaClass, m.Model)):
    class Meta(object):
        abstract = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from functools import wraps
import logging

from django.conf import settings
from django.conf.urls import include, url
from django.http import response
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
import requests
from wechatpy.constants import WeChatErrorCode
from wechatpy.exceptions import WeChatClientException

from .. import settings as wechat_settings
from ..utils.web import auto_response


def patch_request(request, appname=None, cls=None, **kwargs):
    from ..models import WeChatInfo

    cls = cls or WeChatInfo
    appname = appname or request.wechat.appname
    wechat = cls(_appname=appname, _request=request)
    for key, value in kwargs.items():
        setattr(wechat, "_" + key, value)
    request.wechat = wechat
    return request


class WeChatSite(object):
    name = "wechat_django"
    _registered_viewsets = []

    def register(self, cls):
        viewset = cls(self)
        self._registered_viewsets.append(viewset)

    def unregister(self, cls):
        viewset = self.get_registered(cls)
        viewset and self._registered_viewsets.remove(viewset)

    def get_registered(self, cls):
        for viewset in self._registered_viewsets:
            if isinstance(viewset, cls):
                return viewset

    def get_urls(self):
        return [
            url(r"^(?P<appname>[-_a-zA-Z\d]+)/", include(viewset.urls))
            for viewset in self._registered_viewsets
        ]

    @property
    def urls(self):
        return self.get_urls(), "wechat_django", self.name

    @property
    def app_queryset(self):
        """本站点能查询到的所有app
        :rtype: wechat_django.models.app.WeChatAppQuerySet
        """
        from ..models import WeChatApp

        return WeChatApp.objects.get_queryset()


class BaseWeChatViewSet(object):
    def __init__(self, site):
        """:type site: WeChatSite"""
        self.site = site

    def get_urls(self):
        raise NotImplementedError()

    @property
    def urls(self):
        return self.get_urls()


class WeChatViewSet(BaseWeChatViewSet):
    def wechat_view(self, view, methods=None):
        """通过wechat_view装饰的view
        request变为``wechat_django.requests.WeChatHttpRequest``实例
        request中将带有``wechat_django.models.WeChatInfo``类型的wechat属性
        """
        methods = methods or ("GET",)

        @wraps(view)
        def decorated_view(request, appname, *args, **kwargs):
            # 只允许queryset内的appname访问本站点
            app = get_object_or_404(self.site.app_queryset, name=appname)
            request = patch_request(request, appname, _app=app)
            resp = view(request, *args, **kwargs)
            return auto_response(resp)

        rv = csrf_exempt(decorated_view)
        rv = require_http_methods(methods)(rv)

        return rv

    def get_urls(self):
        from ..handler import handler

        return [
            url(
                r"^materials/(?P<media_id>[-_a-zA-Z\d]+)$",
                self.wechat_view(self.material_proxy),
                name="material_proxy"
            ),
            url(
                r"^$",
                self.wechat_view(handler, ("GET", "POST")),
                name="handler"
            )
        ]

    def material_proxy(self, request, media_id):
        """代理下载微信的素材"""
        app = request.wechat.app
        try:
            resp = app.client.material.get(media_id)
        except WeChatClientException as e:
            if e.errcode == WeChatErrorCode.INVALID_MEDIA_ID:
                return response.HttpResponseNotFound()
            app.logger("site").warning(
                "an exception occurred when download material",
                exc_info=True)
            return response.HttpResponseServerError()
        if not isinstance(resp, requests.Response):
            # 暂时只处理image和voice
            return response.HttpResponseNotFound()

        rv = response.FileResponse(resp.content)
        for k, v in resp.headers.items():
            if k.lower().startswith("content-"):
                rv[k] = v
        return rv


default_site = WeChatSite()
"""默认微信站点,适用于一般状况"""
default_site.register(WeChatViewSet)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m, transaction
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _

from ..utils.func import next_chunk
from . import appmethod, WeChatApp, WeChatModel, WeChatUser


class UserTag(WeChatModel):
    SYS_TAGS = (0, 1, 2)

    _id = m.AutoField(primary_key=True)

    app = m.ForeignKey(
        WeChatApp, related_name="user_tags", null=False, editable=False,
        on_delete=m.CASCADE)
    id = m.IntegerField(_("tag id"))
    name = m.CharField(_("tag name"), max_length=30)
    users = m.ManyToManyField(WeChatUser, related_name="tags")

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)

    class Meta(object):
        verbose_name = _("user tag")
        verbose_name_plural = _("user tags")

        unique_together = (("app", "id"), )
        index_together = (("app", "name"), )
        ordering = ("app", "id")

    @property
    def sys_tag(self):
        return self.id in self.SYS_TAGS

    def __init__(self, *args, **kwargs):
        tag_local = kwargs.pop("_tag_local", None)
        super(UserTag, self).__init__(*args, **kwargs)
        self._tag_local = tag_local

    @classmethod
    @appmethod("sync_usertags")
    def sync(cls, app):
        rv = []
        with transaction.atomic():
            db_tags = app.user_tags.all()
            db_tag_ids = {tag.id for tag in db_tags}
            tags = app.client.group.get()

            # 处理已经移除的tags
            deleted_tag_ids = db_tag_ids.difference(
                {tag["id"] for tag in tags})
            app.user_tags.filter(id__in=deleted_tag_ids).delete()

            for tag in tags:
                data = dict(id=tag["id"], name=tag["name"])
                if tag["id"] not in db_tag_ids:
                    db_tag = app.user_tags.create(_tag_local=True, **data)
                else:
                    db_tag = (app.user_tags.filter(id=tag["id"])
                        .update(name=tag["name"]))
                rv.append(db_tag)
            # 存在一个问题
            # 如果先同步了用户 id 101的tag名为a 假设u用户被打了tag a
            # 随后在其他平台移除了tag a 新建了tag b tag b的id恰好为101
            # 重新同步标签 u用户外键依然保留 而tag a已经被删除了
            # 这时这名用户会被认为是tag b的用户
            return rv

    def sync_users(self, detail=True):
        """同步该标签下的所有用户
        :param detail: 是否同步用户详情
        """
        rv = []
        iterator = self.app.client.tag.iter_tag_users(self.id)
        for openids in next_chunk(iterator):
            users = WeChatUser.upsert_users(self.app, openids, detail)
            rv.extend(users)
        return users

    def save(self, *args, **kwargs):
        # 保存之前 先创建标签
        if not getattr(self, "_tag_local", False):
            if self.sys_tag:
                raise ValueError(_("can not edit a system tag"))

            if self.id:
                # TODO: 未修改标签不要更新
                self.app.client.tag.update(self.id, self.name)
            else:
                data = self.app.client.tag.create(self.name)
                self.id = data["id"]
        return super(UserTag, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        take = 50
        limit = 100000
        count = self.users.count()
        if count > limit:
            # 当count 当于100,000时 要先untag users
            diff = count - limit
            for i in range(0, diff, take):
                # 每次只能至多untag 50条记录
                users = self.users.all()[:50]
                self.users.remove(users)

        self.app.client.tag.delete(self.id)
        return super(UserTag, self).delete(*args, **kwargs)

    def __str__(self):
        return self.name


@receiver(m.signals.m2m_changed, sender=UserTag.users.through)
def tag_user_changed(sender, instance, action, *args, **kwargs):
    """给某个标签添加多个用户"""
    if not kwargs["reverse"] and not getattr(instance, "_tag_local", False):
        users = WeChatUser.objects.filter(id__in=kwargs["pk_set"]).all()
        openids = [user.openid for user in users]
        client = instance.app.client.tag
        if action == "pre_add":
            client.tag_user(instance.id, openids)
        elif action == "pre_remove":
            client.untag_user(instance.id, openids)


@receiver(m.signals.m2m_changed, sender=WeChatUser.tags.through)
def user_tag_changed(sender, instance, action, *args, **kwargs):
    """给某个用户添加多个标签"""
    if kwargs["reverse"] and not getattr(instance, "_tag_local", False):
        tags = UserTag.objects.filter(_id__in=kwargs["pk_set"]).all()
        tag_ids = [tag.id for tag in tags]
        client = instance.app.client.tag
        for tag in tags:
            if action == "pre_add":
                client.tag_user(tag.id, instance.openid)
            elif action == "pre_remove":
                client.untag_user(tag.id, instance.openid)



# -*- coding: utf-8 -*-

"""公众号能力"""

from __future__ import unicode_literals

from functools import wraps

from ..constants import AppType


def ability(func):
    @wraps(func)
    def decorated_func(self):
        return bool(func(self))
    return property(decorated_func)


def apptypes(*args, **kwargs):
    def decorator(func):
        @wraps(func)
        def decorated_func(self):
            if kwargs:
                rv = self.app.type not in kwargs["excludes"]
            else:
                rv = self.app.type in args
            return rv and func(self)
        return decorated_func
    return decorator


class Abilities(object):
    """微信号能力"""

    def __init__(self, app=None):
        """:type app: wechat_django.models.self.app"""
        self._app = app

    @ability
    def authed(self):
        """已认证"""
        return self.app.Flag.UNAUTH ^ (self.app.flags & self.app.Flag.UNAUTH)

    @ability
    def api(self):
        """是否可调用微信api(可换取accesstoken)"""
        return self.app.appsecret

    @ability
    def interactable(self):
        """是否可与微信进行消息交互"""
        rv = self.app.token
        if self.app.encoding_mode == self.app.EncodingMode.SAFE:
            rv = rv and self.app.encoding_aes_key
        return rv

    @ability
    @apptypes(AppType.SERVICEAPP)
    def oauth(self):
        """是否可进行网页授权"""
        return self.authed and self.api

    @ability
    @apptypes(AppType.SERVICEAPP, AppType.SUBSCRIBEAPP)
    def menus(self):
        """是否可配置菜单"""
        return self.authed and self.api

    @ability
    @apptypes(AppType.SERVICEAPP, AppType.MINIPROGRAM)
    def template(self):
        """发送模板消息"""
        return self.authed and self.api

    @ability
    @apptypes(AppType.SUBSCRIBEAPP, AppType.SERVICEAPP)
    def user_manager(self):
        """管理用户能力"""
        return self.authed and self.api

    @ability
    @apptypes(AppType.SUBSCRIBEAPP, AppType.SERVICEAPP)
    def material(self):
        """管理用户能力"""
        return self.api

    @ability
    @apptypes(excludes=(AppType.SUBSCRIBEAPP,))
    def pay(self):
        """微信支付能力"""
        return self.authed and self.app.pay

    @property
    def app(self):
        return self._app



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re

from django.utils.translation import ugettext_lazy as _


def enum2choices(enum):
    pattern = re.compile(r"^[A-Z][A-Z_\d]+$")
    return tuple(
        (getattr(enum, key), _(key))
        for key in dir(enum)
        if re.match(pattern, key)
    )


def model_fields(model, excludes=None):
    excludes = excludes or set()
    return set(map(lambda o: o.name, model._meta.fields)).difference(excludes)



# -*- coding: utf-8 -*-

"""微信消息处理器"""

from __future__ import unicode_literals

from contextlib import contextmanager
from functools import wraps
import logging
import time

from django.core.cache import cache
from django.http import response
from django.utils.datastructures import MultiValueDictKeyError
from django.views import View
import six
from wechatpy import replies
from wechatpy.exceptions import InvalidSignatureException
from wechatpy.utils import check_signature
import xmltodict

from . import settings
from .exceptions import BadMessageRequest, MessageHandleError

__all__ = ("handler", "Handler", "message_handler")


class Handler(View):
    def dispatch(self, request):
        """
        :type request: wechat_django.requests.WeChatMessageRequest
        """
        from .models import MessageHandler

        if not request.wechat.app.abilities.interactable:
            return response.HttpResponseNotFound()
        log = MessageHandler.handlerlog(request)
        try:
            self._verify(request)
            resp = super(Handler, self).dispatch(request)
            if not isinstance(resp, response.HttpResponseNotFound):
                log(logging.DEBUG, "receive a message")
            return resp
        except MultiValueDictKeyError:
            log(logging.WARNING, "bad request args", exc_info=True)
            return response.HttpResponseBadRequest()
        except BadMessageRequest:
            log(logging.WARNING, "bad request", exc_info=True)
            return response.HttpResponseBadRequest()
        except InvalidSignatureException:
            log(logging.WARNING, "invalid signature", exc_info=True)
            return response.HttpResponseBadRequest()
        except xmltodict.expat.ExpatError:
            log(logging.WARNING, "deserialize message failed", exc_info=True)
            return response.HttpResponseBadRequest()
        except MessageHandleError:
            log(logging.WARNING, "handle message failed", exc_info=True)
            return ""
        except:
            log(logging.ERROR, "an unexcepted error occurred", exc_info=True)
            return ""

    def get(self, request):
        return request.GET["echostr"]

    def post(self, request):
        from .models import WeChatMessageInfo
        from .sites.wechat import patch_request

        request = patch_request(request, cls=WeChatMessageInfo)
        reply = self._handle(request.wechat)
        if reply:
            xml = reply.render()
            if request.wechat.app.crypto:
                xml = request.wechat.app.crypto.encrypt_message(
                    xml, request.GET["nonce"], request.GET["timestamp"])
            return response.HttpResponse(xml, content_type="text/xml")
        return ""

    def _verify(self, request):
        """检验请求"""
        try:
            timestamp = int(request.GET["timestamp"])
        except ValueError:
            raise BadMessageRequest("invalid timestamp")
        sign = request.GET["signature"]
        nonce = request.GET["nonce"]

        time_diff = int(timestamp) - time.time()
        # 检查timestamp
        if abs(time_diff) > settings.MESSAGETIMEOFFSET:
            raise BadMessageRequest("invalid time")

        # 防重放检查及签名检查
        with self._no_repeat_nonces(sign, nonce, time_diff):
            check_signature(
                request.wechat.app.token,
                sign,
                timestamp,
                nonce
            )

    @contextmanager
    def _no_repeat_nonces(self, sign, nonce, time_diff):
        """nonce防重放"""
        nonce_key = "wx:m:n:{0}".format(sign)
        if settings.MESSAGENOREPEATNONCE:
            if cache.get(nonce_key) == nonce:
                raise BadMessageRequest("repeat nonce string")
            try:
                yield
                expires = settings.MESSAGETIMEOFFSET + time_diff
                cache.set(nonce_key, nonce, expires)
            finally:
                pass
        else:
            yield

    def _handle(self, message_info):
        """处理消息"""
        from .models import MessageHandler, MessageLog

        handlers = MessageHandler.matches(message_info)
        if not handlers:
            return None

        handler = handlers[0]
        reply = handler.reply(message_info)
        if handler.log_message or message_info.app.log_message:
            MessageLog.from_message_info(message_info)
        if not reply or isinstance(reply, replies.EmptyReply):
            return None
        return reply


def message_handler(names_or_func=None):
    """自定义回复业务需加装该装饰器
    被装饰的自定义业务接收一个``wechat_django.models.WeChatMessageInfo``对象
    并且返回一个``wechatpy.replies.BaseReply``对象

    :param names_or_func: 允许使用该message_handler的appname 不填所有均允许
    :type names_or_func: str or list or tuple or callable

        @message_handler
        def custom_business(message):
            user = message.user
            # ...
            return TextReply("hello")

        @message_handler(("app_a", "app_b"))
        def app_ab_only_business(message):
            # ...
    """
    def decorator(view_func):
        @wraps(view_func)
        def decorated_view(message):
            return view_func(message)
        decorated_view.message_handler = names or True

        return decorated_view

    if isinstance(names_or_func, six.text_type):
        names = [names_or_func]
    elif callable(names_or_func):
        names = None
        return decorator(names_or_func)

    return decorator


handler = Handler.as_view()



# -*- coding: utf-8 -*-

"""
微信相关的请求类声明,这些请求类中,包含一个WeChatInfo对象wechat
实际上这些请求类没有真正被使用
但是可以通过docstring声明传参是这些请求类,让你的IDE更加智能地提示代码
"""

from __future__ import unicode_literals

from django.http.request import HttpRequest


class WeChatHttpRequest(HttpRequest):
    @property
    def wechat(self):
        """:rtype: wechat_django.models.WeChatInfo"""
        pass


class WeChatMessageRequest(HttpRequest):
    @property
    def wechat(self):
        """:rtype: wechat_django.models.WeChatMessageInfo"""
        pass


class WeChatOAuthRequest(HttpRequest):
    @property
    def wechat(self):
        """:rtype: wechat_django.models.WeChatOAuthInfo"""
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class MsgType(object):
    TEXT = "text"
    IMAGE = "image"
    VOICE = "voice"
    VIDEO = "video"


class MsgLogFlag(object):
    LOG_MESSAGE = 0x01
    LOG_REPLY = 0x02


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import random

from django.db import models as m, transaction
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from wechatpy.exceptions import WeChatClientException

from ..exceptions import MessageHandleError
from ..utils.model import enum2choices
from ..utils.web import get_ip
from . import appmethod, MsgLogFlag, WeChatApp, WeChatModel


class MessageHandlerManager(m.Manager):
    def create_handler(self, rules=None, replies=None, **kwargs):
        """:rtype: wechat_django.models.MessageHandler"""
        handler = self.create(**kwargs)
        if rules:
            for rule in rules:
                rule.handler = handler
            handler.rules.bulk_create(rules)
        if replies:
            for reply in replies:
                reply.handler = handler
            handler.replies.bulk_create(replies)
        return handler


class MessageHandler(WeChatModel):
    class Source(object):
        SELF = 0  # 自己的后台
        MENU = 1  # 菜单
        MP = 2  # 微信后台

    class ReplyStrategy(object):
        REPLYALL = "reply_all"
        RANDOM = "random_one"
        NONE = "none"

    class EventType(object):
        SUBSCRIBE = "subscribe"
        UNSUBSCRIBE = "unsubscribe"
        SCAN = "SCAN"
        LOCATION = "LOCATION"
        CLICK = "CLICK"
        VIEW = "VIEW"

    class Flag(object):
        TERMINALONEXCEPTION = 0x01 # 当回复全部时 一个回复发生异常 中断其他回复

    app = m.ForeignKey(
        WeChatApp, on_delete=m.CASCADE, related_name="message_handlers",
        null=False, editable=False)
    name = m.CharField(_("name"), max_length=60)
    src = m.PositiveSmallIntegerField(choices=(
        (Source.MP, "wechat"),
        (Source.SELF, "self"),
        (Source.MENU, "menu")
    ), default=Source.SELF, editable=False)
    strategy = m.CharField(_("strategy"), max_length=10,
        choices=enum2choices(ReplyStrategy), default=ReplyStrategy.REPLYALL)

    flags = m.IntegerField(_("flags"), default=False)

    starts = m.DateTimeField(_("starts"), null=True, blank=True)
    ends = m.DateTimeField(_("ends"), null=True, blank=True)
    enabled = m.BooleanField(_("enabled"), null=False, default=True)

    weight = m.IntegerField(_("weight"), default=0, null=False)
    created_at = m.DateTimeField(_("created_at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated_at"), auto_now=True)

    objects = MessageHandlerManager()

    class Meta:
        verbose_name = _("message handler")
        verbose_name_plural = _("message handlers")

        ordering = ("-weight", "-created_at", "-id")
        index_together = (
            ("app", "weight", "created_at"),
        )

    @property
    def log_message(self):
        return bool(self.flags & MsgLogFlag.LOG_MESSAGE)

    @property
    def available(self):
        if not self.enabled:
            return False
        now = timezone.now()
        if self.starts and self.starts > now:
            return False
        if self.ends and self.ends < now:
            return False
        return True

    @staticmethod
    def matches(message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        handlers = message_info.app.message_handlers\
            .prefetch_related("rules").all()
        for handler in handlers:
            if handler.is_match(message_info):
                return (handler, )

    def is_match(self, message_info):
        if self.available:
            for rule in self.rules.all():
                if rule.match(message_info):
                    return self

    def reply(self, message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        :rtype: wechatpy.replies.BaseReply
        """
        reply = ""
        if self.strategy == self.ReplyStrategy.NONE:
            pass
        else:
            replies = list(self.replies.all())
            if not replies:
                pass
            elif self.strategy == self.ReplyStrategy.REPLYALL:
                for reply in replies[:-1]:
                    try:
                        reply.send(message_info)
                    except Exception as e:
                        # 发送异常 继续处理其他程序
                        log = self.handlerlog(message_info.request)
                        msg = "an unexcepted error occurred when send msg"
                        level = logging.WARNING\
                            if isinstance(e, WeChatClientException)\
                            else logging.ERROR
                        log(level, msg, exc_info=True)
                reply = replies[-1]
            elif self.strategy == self.ReplyStrategy.RANDOM:
                reply = random.choice(replies)
            else:
                raise MessageHandleError("incorrect reply strategy")
        return reply and reply.reply(message_info)

    @classmethod
    @appmethod("sync_message_handlers")
    def sync(cls, app):
        from . import Reply, Rule
        resp = app.client.message.get_autoreply_info()

        # 处理自动回复
        handlers = []

        # 成功后移除之前的自动回复并保存新加入的自动回复
        with transaction.atomic():
            app.message_handlers.filter(
                src=MessageHandler.Source.MP
            ).delete()

            if resp.get("message_default_autoreply_info"):
                # 自动回复
                handler = app.message_handlers.create_handler(
                    name="微信配置自动回复",
                    src=MessageHandler.Source.MP,
                    enabled=bool(resp.get("is_autoreply_open")),
                    created_at=timezone.datetime.fromtimestamp(0),
                    rules=[Rule(type=Rule.Type.ALL)],
                    replies=[
                        Reply.from_mp(
                            app, resp["message_default_autoreply_info"])
                    ]
                )
                handlers.append(handler)

            if (resp.get("keyword_autoreply_info")
                and resp["keyword_autoreply_info"].get("list")):
                handlers_list = resp["keyword_autoreply_info"]["list"][::-1]
                handlers.extend(
                    MessageHandler.from_mp(app, handler)
                    for handler in handlers_list
                )

            if resp.get("add_friend_autoreply_info"):
                # 关注回复
                handler = app.message_handlers.create_handler(
                    name="微信配置关注回复",
                    src=MessageHandler.Source.MP,
                    enabled=bool(resp.get("is_add_friend_reply_open")),
                    created_at=timezone.datetime.fromtimestamp(0),
                    rules=[Rule(
                        type=Rule.Type.EVENT,
                        event=cls.EventType.SUBSCRIBE
                    )],
                    replies=[
                        Reply.from_mp(app, resp["add_friend_autoreply_info"])
                    ]
                )
                handlers.append(handler)

            return handlers

    @classmethod
    def from_mp(cls, app, handler):
        from . import Reply, Rule
        return app.message_handlers.create_handler(
            name=handler["rule_name"],
            src=MessageHandler.Source.MP,
            created_at=timezone.datetime.fromtimestamp(handler["create_time"]),
            strategy=handler["reply_mode"],
            rules=[
                Rule.from_mp(rule)
                for rule in handler["keyword_list_info"][::-1]
            ],
            replies=[
                Reply.from_mp(app, reply)
                for reply in handler["reply_list_info"][::-1]
            ]
        )

    @classmethod
    def from_menu(cls, menu, data):
        """
        :type menu: wechat_django.models.Menu
        """
        from . import Reply, Rule
        return menu.app.message_handlers.create_handler(
            name="菜单[{0}]事件".format(data["name"]),
            src=cls.Source.MENU,
            rules=[Rule(
                type=Rule.Type.EVENTKEY,
                event=cls.EventType.CLICK,
                key=menu.content["key"]
            )],
            replies=[Reply.from_menu(menu, data)]
        )

    @classmethod
    def handlerlog(cls, request):
        logger = request.wechat.app.logger("handler")
        args = dict(
            params=request.GET,
            body=request.body,
            ip=get_ip(request)
        )
        s = "%s - {0}".format(args)
        return lambda lvl, msg, **kwargs: logger.log(lvl, s % msg, **kwargs)

    def __str__(self):
        return "{0}".format(self.name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from contextlib import contextmanager
from functools import wraps

from django.conf.urls import include, url
from django.contrib import admin
from django.http import response
from django.template.response import SimpleTemplateResponse
from django.urls import NoReverseMatch, resolve, Resolver404, reverse
from django.utils.http import urlencode
from django.utils.translation import ugettext_lazy as _
from object_tool import CustomObjectToolAdminSiteMixin
from six.moves.urllib.parse import urlparse

from ..admin.base import registered_admins, WeChatModelAdmin
from ..admin.utils import get_request_params
from ..models import WeChatApp
from ..models.permission import get_user_permissions
from .wechat import default_site as default_wechat_site


def wechat_admin_view(view, site):
    """装饰WeChatAdmin中的view
    在请求上附上WeChatApp实例
    并在响应的模板上附上app,app_id等context
    """
    def correct_url(request, object_id):
        """修正请求url"""
        resolved_url = resolve(request.path_info)
        url_name = resolved_url.url_name
        url_kwargs = resolved_url.kwargs

        if object_id:
            # 对于只有object_id没有app_id的 重定向到有app_id的url
            modeladmin = view.__self__
            try:
                obj = modeladmin.model.objects.get(pk=object_id)
            except modeladmin.model.DoesNotExist:
                return response.HttpResponseBadRequest()

            url_kwargs.update(
                wechat_app_id=obj.app_id,
                object_id=object_id
            )
            return response.HttpResponseRedirect(
                reverse("admin:" + url_name, kwargs=url_kwargs), status=307
            )
        else:
            # 对于没有app_id的listview请求 优先取referrer的app_id
            referrer = request.META.get("HTTP_REFERER", "")
            path_info = urlparse(referrer).path
            try:
                app_id = resolve(path_info).kwargs["wechat_app_id"]
            except (KeyError, Resolver404):
                return response.HttpResponseNotFound()

            url_kwargs.update(wechat_app_id=app_id)
            resp = response.HttpResponseRedirect(
                reverse("admin:" + url_name, kwargs=url_kwargs)
            )
            resp.status_code = 307
            return resp

    @wraps(view)
    def decorated_func(request, *args, **kwargs):
        app_id = kwargs.pop("wechat_app_id", None)
        object_id = kwargs.get("object_id", None)
        if not app_id:
            return correct_url(request, object_id)

        extra_context = kwargs.pop("extra_context", None) or {}
        try:
            app = site.wechat_site.app_queryset.get(id=app_id)
        except WeChatApp.DoesNotExist:
            return response.HttpResponseNotFound()

        # 附上app
        request.app = app
        request.app_id = app_id
        # 增加模板context
        extra_context.update(
            wechat_app=app,
            wechat_app_id=app_id
        )
        kwargs["extra_context"] = extra_context
        return view(request, *args, **kwargs)

    return decorated_func


class WeChatAdminSiteMixin(CustomObjectToolAdminSiteMixin):
    """AdminSiteMixin 自定义后台需要包含微信相关功能时 需要将本Mixin混入"""
    _default_wechat_site = default_wechat_site

    def __init__(self, *args, **kwargs):
        super(WeChatAdminSiteMixin, self).__init__(*args, **kwargs)
        for model, admin_class in registered_admins:
            self.register(model, admin_class)

    @property
    def wechat_site(self):
        """默认微信站点,在获取app queryset时与做url reverse时需使用

        :rtype: wechat_django.WeChatSite
        """
        return self._default_wechat_site

    @contextmanager
    def _unregister_wechatadmins(self):
        """暂时取消注册wechat-django相关的modeladmin以进行一些操作"""
        for model, model_admin in registered_admins:
            self.unregister(model)
        yield
        for model, model_admin in registered_admins:
            self.register(model, model_admin)

    def _iter_wechatadmins(self):
        """枚举所有的wechat-django modeladmin实例"""
        for model, model_admin in registered_admins:
            yield model, self._registry[model]

    def admin_view(self, view, cacheable=False):
        model_admin = getattr(view, "__self__", None)
        # 对于wechat-django相关的admin view 做一下装饰
        if isinstance(model_admin, WeChatModelAdmin):
            view = wechat_admin_view(view, self)
            model_admin.wechat_site = self.wechat_site

        return super(WeChatAdminSiteMixin, self).admin_view(view, cacheable)

    def get_urls(self):
        with self._unregister_wechatadmins():
            urlpatterns = super(WeChatAdminSiteMixin, self).get_urls()

        # 重新注册wechat_django相关的modeladmin
        for model, model_admin in self._iter_wechatadmins():
            info = model._meta.app_label, model._meta.model_name
            urlpatterns += [
                url(
                    r"^%s/(?:(?P<wechat_app_id>\d+)/)?%s/" % info,
                    include(model_admin.urls)
                ),
            ]

        urlpatterns += [
            url(
                r"(?P<app_label>wechat_django)/apps/(?P<wechat_app_id>\d+)/$",
                self.admin_view(wechat_admin_view(self.wechat_index, self)),
                name="wechat_funcs_list"
            )
        ]

        return urlpatterns

    def _build_app_dict(self, request, label=None):
        app_id = getattr(request, "app_id", None)
        if app_id:
            # 微信相关模块
            rv = self._build_wechat_func_dict(request)
            
            if not label:
                # 我也不记得为啥要加这个了..
                return {WeChatApp._meta.app_label: rv}
        else:
            with self._unregister_wechatadmins():
                rv = super(WeChatAdminSiteMixin, self)._build_app_dict(request, label)

            # 首页, 增加app列表
            if not label:
                app_dict = self._build_wechat_app_dict(request)
                if app_dict["has_module_perms"]:
                    rv["wechat_django_apps"] = app_dict

        return rv

    def _build_wechat_app_dict(self, request):
        """构建wechat apps列表"""
        app_label = WeChatApp._meta.app_label

        # 过滤有权限的app
        query = self.wechat_site.app_queryset
        if not request.user.is_superuser:
            perms = get_user_permissions(request.user)
            allowed_apps = {
                k for k, ps in perms.items() if ps != {"manage"}
            }
            query = query.filter(name__in=allowed_apps)
        apps = query.all()

        # 构建app_dict
        app_perms = [
            dict(
                name=str(app),
                object_name=app.name,
                perms=dict(
                    change=True,
                ),
                admin_url=reverse(
                    "admin:wechat_funcs_list",
                    current_app=self.name,
                    kwargs=dict(
                        wechat_app_id=app.id,
                        app_label=app_label
                    )
                )
            )
            for app in apps
        ]
        return {
            'name': WeChatApp._meta.verbose_name_plural,
            'app_label': app_label,
            'app_url': "#",
            'has_module_perms': bool(app_perms),
            'models': app_perms,
        }

    def _build_wechat_func_dict(self, request):
        """构建某一公众号下功能列表"""
        has_module_perms = get_user_permissions(
            request.user, request.app, exclude_manage=True, exclude_sub=True)
        if not has_module_perms:
            return None

        app_id = request.app_id
        app_label = WeChatApp._meta.app_label
        rv = dict(
            name=str(request.app),
            app_label=app_label,
            app_url=reverse(
                "admin:wechat_funcs_list",
                current_app=self.name,
                kwargs=dict(
                    wechat_app_id=app_id,
                    app_label=app_label
                )
            ),
            has_module_perms=True,
            models=[]
        )

        for model, model_admin in self._iter_wechatadmins():
            info = (model._meta.app_label, model._meta.model_name)
            perms = model_admin.get_model_perms(request)
            if not perms:
                continue
            model_dict = {
                "name": model._meta.verbose_name_plural,
                "object_name": model._meta.object_name,
                "perms": perms,
            }
            if perms.get("change"):
                try:
                    model_dict["admin_url"] = reverse(
                        "admin:%s_%s_changelist" % info,
                        current_app=self.name,
                        kwargs=dict(wechat_app_id=app_id)
                    )
                except NoReverseMatch:
                    pass
            if perms.get("add"):
                try:
                    model_dict['add_url'] = reverse(
                        "admin:%s_%s_add" % info,
                        current_app=self.name,
                        kwargs=dict(wechat_app_id=app_id)
                    )
                except NoReverseMatch:
                    pass
            rv["models"].append(model_dict)

        return rv

    def wechat_index(self, request, *args, **kwargs):
        """某个公众号后台管理首页"""
        return super(WeChatAdminSiteMixin, self).app_index(
            request, *args, **kwargs)


class WeChatAdminSite(WeChatAdminSiteMixin, admin.AdminSite):
    def from_site(self, site):
        self._registry.update(site._registry)
        for model, model_admin in self._registry.items():
            model_admin.admin_site = self


def patch_admin():
    """用当前wechat-django默认的adminsite替代django自带的默认adminsite"""
    default_site.from_site(admin.site)
    setattr(admin.sites, "site", default_site)
    setattr(admin, "site", default_site)


default_site = WeChatAdminSite()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from django.http.response import HttpResponse, HttpResponseNotFound
from django.shortcuts import redirect
from django.utils.http import urlencode
from django.views import View
import six
from wechatpy import WeChatOAuth as _WeChatOAuthClient, WeChatOAuthException

from .constants import WeChatSNSScope
from .utils.web import auto_response, get_params

__all__ = (
    "wechat_auth", "WeChatOAuthClient", "WeChatOAuthView", "WeChatSNSScope")


class WeChatOAuthClient(_WeChatOAuthClient):
    OAUTH_URL = "https://open.weixin.qq.com/connect/oauth2/authorize"
    QRCONNECT_URL = "https://open.weixin.qq.com/connect/qrconnect"

    def __init__(self, app):
        if app.configurations.get("OAUTH_URL"):
            self.OAUTH_URL = app.configurations["OAUTH_URL"]
        super(WeChatOAuthClient, self).__init__(app.appid, app.appsecret, "")

    def authorize_url(self, redirect_uri, scope=WeChatSNSScope.BASE, state=""):
        return self.OAUTH_URL + "?" + urlencode(dict(
            appid=self.app_id,
            redirect_uri=redirect_uri,
            response_type="code",
            scope=scope,
            state=state
        )) + "#wechat_redirect"

    def qrconnect_url(self, redirect_uri, state=""):
        return self.QRCONNECT_URL + "?" + urlencode(dict(
            appid=self.app_id,
            redirect_uri=redirect_uri,
            response_type="code",
            scope="snsapi_login",
            state=state
        )) + "#wechat_redirect"


class wechat_auth(object):
    def __init__(
        self, appname, scope=None, redirect_uri=None, required=True,
        response=None, state=""
    ):
        """微信网页授权
        :param appname: WeChatApp的name
        :param scope: 默认WeChatSNSScope.BASE, 可选WeChatSNSScope.USERINFO
        :type scope: str or iterable
        :param redirect_uri: 未授权时的重定向地址 当未设置response时将自动执行授权
                            当ajax请求时默认取referrer 否则取当前地址
                            注意 请不要在地址上带有code及state参数 否则可能引发问题
        :param state: 授权时需要携带的state
        :type state: str
                     or Callable[
                        [
                            django.http.request.HttpRequest,
                            *args,
                            **kwargs
                        ],
                        str
                     ]
        :param required: 真值必须授权 否则不授权亦可继续访问(只检查session)
        :param response: 未授权的返回 接受一个
        :type response: django.http.response.HttpResponse
                        or Callable[
                            [
                                django.http.request.HttpRequest,
                                *args,
                                **kwargs
                            ],
                            django.http.response.HttpResponse
                        ]

        使用示例:

            @wechat_auth("your_app_name")
            def your_view(request, *args, **kwargs):
                # request是一个``wechat_django.requests.WeChatOAuthRequest对象``
                user = request.wechat.user
        """
        scope = scope or (WeChatSNSScope.BASE,)
        if isinstance(scope, six.text_type):
            scope = (scope,)

        assert appname and isinstance(appname, six.text_type),\
            "incorrect appname"
        assert (
            response is None or callable(response)
            or isinstance(response, HttpResponse)
        ), "incorrect response param"
        for s in scope:
            assert s in (WeChatSNSScope.BASE, WeChatSNSScope.USERINFO),\
                "incorrect scope"

        self.appname = appname
        self.scope = scope
        self._redirect_uri = redirect_uri
        self.required = required
        self.response = response
        self.state = state

    def redirect_uri(self, request):
        return request.build_absolute_uri(
            self._redirect_uri
            or (request.is_ajax() and request.META.get("HTTP_REFERER"))
            or request.build_absolute_uri()
        )

    def __call__(self, view):
        return WeChatOAuthView(self, view)

    def __str__(self):
        return "<wechat_auth appname: {appname} scope: {scope}>".format(
            appname=self.appname,
            scope=self.scope
        )


class WeChatOAuthView(View):
    @property
    def oauth_info(self):
        """:rtype: wechat_django.oauth.wechat_auth"""
        if not self._oauth_info:
            self._oauth_info = wechat_auth(
                self.appname, scope=self.scope, state=self.state,
                redirect_uri=self.redirect_uri, required=self.required,
                response=self.response)
        return self._oauth_info

    #region class based view properties
    appname = None
    """
    微信appname 必填
    """

    scope = (WeChatSNSScope.BASE,)
    """
    :type: str or iterable
    微信授权的scope 默认WeChatSNSScope.BASE
    """

    redirect_uri = None
    """
    未授权时的重定向地址 当未设置response时将自动执行授权
    当ajax请求时默认取referrer 否则取当前地址
    注意 请不要在地址上带有code及state参数 否则可能引发问题
    """

    state = ""
    """
    :type: str or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        str
    ]
    授权时需要携带的state
    """

    required = True
    """真值必须授权 否则不授权亦可继续访问(只检查session)"""

    response = None
    """
    :type: django.http.response.HttpResponse or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        django.http.response.HttpResponse
    ]
    """
    #endregion

    def __init__(self, oauth_info=None, view=None, **kwargs):
        """
        :type oauth_info: wechat_django.oauth.wechat_auth
        """
        self._oauth_info = oauth_info
        if view:
            self.get = view
        super(WeChatOAuthView, self).__init__(**kwargs)

    def dispatch(self, request, *args, **kwargs):
        from .models import WeChatApp

        try:
            self._patch_request(request)
            return self._dispatch(request, *args, **kwargs)
        except WeChatApp.DoesNotExist:
            return HttpResponseNotFound()

    def _dispatch(self, request, *args, **kwargs):
        wechat = request.wechat

        code = get_params(self.request, "code")
        if not wechat.openid and code:
            # 有code 先授权
            try:
                user = self.auth(wechat.app, code)
                wechat._openid = user.openid
                wechat._user = user
                # 用当前url的state替换传入的state
                wechat._state = get_params(request, "state", "")
            except WeChatOAuthException:
                err_msg = "auth code failed: {0}".format(dict(
                    info=wechat,
                    code=code
                ))
                wechat.app.logger("oauth").warning(err_msg, exc_info=True)

        # 没有openid 响应未授权
        if self.oauth_info.required and not wechat.openid:
            return self.unauthorization_response(request, *args, **kwargs)

        response = super(WeChatOAuthView, self).dispatch(
            request, *args, **kwargs)
        response = auto_response(response)
        wechat.openid and response.set_signed_cookie(
            wechat.session_key, wechat.openid)
        return response

    def unauthorization_response(self, request, *args, **kwargs):
        """未授权的响应"""
        response = self.oauth_info.response
        if response and callable(response):
            response = response(request, *args, **kwargs)
        elif not response:
            oauth_uri = request.wechat.oauth_uri
            response = redirect(oauth_uri, permanent=False)
        return response

    def auth(self, app, code):
        return app.auth(code, self.oauth_info.scope)[0]

    def _patch_request(self, request, *args, **kwargs):
        from .models import WeChatOAuthInfo
        from .sites.wechat import patch_request

        info = self.oauth_info
        state = info.state(request, *args, **kwargs) if callable(info.state)\
            else info.state
        self.request = patch_request(
            request=request,
            appname=info.appname,
            cls=WeChatOAuthInfo,
            redirect_uri=info.redirect_uri(request),
            scope=info.scope,
            state=state
        )
        self.args = args
        self.kwargs = kwargs

    __call__ = dispatch



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

import six
from wechatpy import WeChatPay as WeChatPayBaseClient

from .base import mock, WeChatPayTestCase


class PayTestCase(WeChatPayTestCase):
    def test_client_init(self):
        """测试WeChatPayClient的构建"""
        pay = self.app.pay
        sub_pay = self.app_sub.pay

        mch_id = pay.mch_id
        api_key = pay.api_key
        sub_mch_id = sub_pay.sub_mch_id
        mch_app_id = sub_pay.mch_app_id

        with mock.patch.object(WeChatPayBaseClient, "__init__"):
            pay.client
            self.assertCallArgsEqual(
                WeChatPayBaseClient.__init__, kwargs=dict(
                    mch_id=mch_id,
                    api_key=api_key,
                    appid=pay.app.appid
                ))

            sub_pay.client
            self.assertCallArgsEqual(
                WeChatPayBaseClient.__init__, kwargs=dict(
                    mch_id=mch_id,
                    api_key=api_key,
                    appid=mch_app_id,
                    sub_appid=sub_pay.app.appid,
                    sub_mch_id=sub_mch_id
                ))

    def test_client_cert(self):
        """测试请求时证书是否正确使用"""
        pay = self.app_nocert.pay
        cert_pay = self.app.pay
        mch_id = cert_pay.mch_id
        api_key = cert_pay.api_key
        mch_cert = cert_pay.mch_cert
        mch_key = cert_pay.mch_key
        if six.PY2:
            mch_cert = str(mch_cert)
            mch_key = str(mch_key)

        that = self
        origin_request = WeChatPayBaseClient._request
        try:
            # 无证书请求时不带证书文件
            def mock_request(self, *args, **kwargs):
                that.assertFalse(self.mch_cert)
                that.assertFalse(self.mch_key)
            WeChatPayBaseClient._request = mock_request
            pay.client.order.query("1")

            # 有证书请求时带证书文件
            def mock_request(self, *args, **kwargs):
                that.assertTrue(self.mch_cert)
                that.assertTrue(self.mch_key)
                with open(self.mch_cert, "rb") as mch_cert_file,\
                    open(self.mch_key, "rb") as mch_key_file:
                    that.assertEqual(mch_cert_file.read(), mch_cert)
                    that.assertEqual(mch_key_file.read(), mch_key)
                return self.mch_cert, self.mch_key
            WeChatPayBaseClient._request = mock_request
            mch_cert_name, mch_key_name = cert_pay.client.order.query("1")

            # 请求结束后已移除证书文件
            self.assertFalse(os.path.exists(mch_cert_name))
            self.assertFalse(os.path.exists(mch_key_name))
        finally:
            WeChatPayBaseClient._request = origin_request



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.urls import reverse
from wechatpy.client import api

from .. import models
from .base import mock, WeChatTestCase


class AdminViewTestCase(WeChatTestCase):
    """测试各admin页面能正常访问"""
    def setUp(self):
        super(AdminViewTestCase, self).setUp()
        user = User.objects.create(
            username="admin", is_superuser=True, is_staff=True)
        user.set_password("123456")
        user.save()
        self.client.login(username="admin", password="123456")

    def test_indexes(self):
        """测试首页"""
        self.assertRequestSuccess(reverse("admin:index"))
        self.assertRequestSuccess(reverse("admin:app_list", kwargs=dict(
            app_label="wechat_django"
        )))
        self.assertRequestSuccess(reverse("admin:wechat_funcs_list", kwargs=dict(
            app_label="wechat_django",
            wechat_app_id=self.app.id
        )))

    def test_article_view(self):
        """测试图文"""
        self.assertModelViewSuccess(
            models.Article, excludes=("history", "delete", "change"),
            status={"add": 403})

        material = models.Material.objects.create(
            app=self.app,
            media_id="media_id"
        )
        article = models.Article.objects.create(
            material=material,
            content="",
            url="url",
            index=0
        )
        self.assertModelViewSuccess(
            models.Article, article.id, status={"add": 403, "delete": 403})

    def test_material_view(self):
        """测试素材"""
        self.assertModelViewSuccess(
            models.Material, excludes=("history", "delete", "change"),
            status={"add": 403})

        material = models.Material.objects.create(
            app=self.app,
            media_id="media_image",
            type=models.Material.Type.IMAGE
        )
        self.assertModelViewSuccess(
            models.Material, material.id, status={"add": 403})

        material = models.Material.objects.create(
            app=self.app,
            media_id="media_video",
            type=models.Material.Type.VIDEO
        )
        self.assertModelViewSuccess(
            models.Material, material.id, status={"add": 403})

        material = models.Material.objects.create(
            app=self.app,
            media_id="media_voice",
            type=models.Material.Type.VOICE
        )
        self.assertModelViewSuccess(
            models.Material, material.id, status={"add": 403})
        
        material = models.Material.objects.create(
            app=self.app,
            media_id="media_new"
        )
        article = models.Article.objects.create(
            material=material,
            content="",
            url="url",
            index=0
        )
        self.assertModelViewSuccess(
            models.Material, material.id, status={"add": 403})

    def test_menu_view(self):
        """测试菜单"""
        self.assertModelViewSuccess(
            models.Menu, excludes=("history", "delete", "change"))
        with mock.patch.object(api.WeChatMenu, "get_menu_info"):
            api.WeChatMenu.get_menu_info.return_value = self.load_data("self_menu_data")
            menus = models.Menu.sync(self.app)
        self.assertModelViewSuccess(models.Menu, menus[0].id)

    def test_messagehandler_view(self):
        """测试消息处理程序"""
        self.assertModelViewSuccess(
            models.MessageHandler, excludes=("history", "delete", "change"))

        handler = models.MessageHandler.objects.create_handler(
            rules=[
                models.Rule(type=models.Rule.Type.ALL)
            ],
            replies=[
                models.Reply(type=models.Reply.MsgType.TEXT, pattern="test")
            ],
            app=self.app
        )
        self.assertModelViewSuccess(models.MessageHandler, handler.id)

    def test_messagelog_view(self):
        """测试消息日志"""
        self.assertModelViewSuccess(
            models.MessageLog, excludes=("history", "delete", "change"),
            status={"add": 403})

        user = models.WeChatUser.objects.create(app=self.app, openid="openid")
        log = models.MessageLog.objects.create(
            app=self.app, user=user, type=models.Rule.ReceiveMsgType.TEXT,
            content={})
        self.assertModelViewSuccess(
            models.MessageLog, log.id, status={"add": 403})

    def test_user_view(self):
        """测试用户"""
        self.assertModelViewSuccess(
            models.WeChatUser, excludes=("history", "delete", "change"),
            status={"add": 403})

        user = models.WeChatUser.objects.create(app=self.app, openid="openid")
        self.assertModelViewSuccess(
            models.WeChatUser, user.id, status={"add": 403, "delete": 403})


    def test_usertag(self):
        """测试用户标签"""
        self.assertModelViewSuccess(
            models.UserTag, excludes=("history", "delete", "change"))

        tag = models.UserTag.objects.create(
            app=self.app, id=101, name="tag", _tag_local=True)
        self.assertModelViewSuccess(models.UserTag, tag.id)

    def test_template(self):
        """测试模板消息后台"""
        self.assertModelViewSuccess(
            models.Template, excludes=("history", "delete", "change"),
            status={"add": 403})

        template = models.Template.objects.create(
            app=self.app, template_id="t", title="title")
        self.assertModelViewSuccess(
            models.Template, template.id, status={"add": 403, "delete": 403})

    def test_app(self):
        """测试公众号"""
        self.assertModelViewSuccess(models.WeChatApp, self.app.id, wechat=False)

    def assertModelViewSuccess(self, model, id=None, excludes=None, status=None, wechat=True):
        excludes = excludes or tuple()
        status = status or {}
        model_name = model._meta.model_name

        def make_url(view, **kwargs):
            wechat and kwargs.update(wechat_app_id=self.app.id)
            pattern = "admin:%s_%s_%s" % ("wechat_django", model_name, view)
            return reverse(pattern, kwargs=kwargs)

        "changelist" not in excludes\
            and self.assertRequestSuccess(
                make_url("changelist"), status_code=status.get("changelist"))
        "add" not in excludes\
            and self.assertRequestSuccess(
                make_url("add"), status_code=status.get("add"))
        # self.assertRequestSuccess(make_url("autocomplete"))
        "history" not in excludes\
            and self.assertRequestSuccess(
                make_url("history", object_id=id),
                status_code=status.get("history"))
        "delete" not in excludes\
            and self.assertRequestSuccess(
                make_url("delete", object_id=id),
                status_code=status.get("delete"))
        "change" not in excludes\
            and self.assertRequestSuccess(
                make_url("change", object_id=id),
                status_code=status.get("change"))

    def assertRequestSuccess(self, url, method="get", **kwargs):
        status_code = kwargs.pop("status_code", None) or 200
        resp = getattr(self.client, method)(url, follow=True, **kwargs)
        self.assertSuccess(resp, status_code)
    
    def assertSuccess(self, resp, status=200):
        self.assertEqual(resp.status_code, status)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.urls import reverse
from django.utils import timezone
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
import object_tool

from ...models import Material
from ..base import RecursiveDeleteActionMixin, WeChatModelAdmin


class MaterialAdmin(RecursiveDeleteActionMixin, WeChatModelAdmin):
    __category__ = "material"
    __model__ = Material

    changelist_object_tools = ("sync", )
    list_display = ("media_id", "alias", "type", "comment", "updatetime")
    list_filter = ("type", )
    search_fields = ("name", "media_id", "comment")

    fields = ("type", "media_id", "alias", "name", "open", "comment")
    readonly_fields = ("type", "media_id", "name", "open", "media_id")

    @mark_safe
    def preview(self, obj):
        if obj.type == Material.Type.IMAGE:
            return '<img src="%s" />' % obj.url
    preview.short_description = _("preview")

    def updatetime(self, obj):
        return (obj.update_time
            and timezone.datetime.fromtimestamp(obj.update_time))
    updatetime.short_description = _("update time")

    @mark_safe
    def open(self, obj):
        blank = True
        if obj.type == Material.Type.NEWS:
            url = "{0}?{1}".format(
                reverse(
                    "admin:wechat_django_article_changelist",
                    kwargs=dict(wechat_app_id=obj.app_id)
                ),
                urlencode(dict(
                    material_id=obj.id
                ))
            )
            blank = False
        elif obj.type == Material.Type.VOICE:
            # 代理下载
            url = obj.app.build_url(
                "material_proxy", kwargs=dict(media_id=obj.media_id))
        else:
            url = obj.url
        return '<a href="{0}" {1}>{2}</a>'.format(
            url, 'target="_blank"' if blank else "", _("open")
        )
    open.short_description = _("open")

    @object_tool.confirm(short_description=_("Sync materials"))
    def sync(self, request, obj=None):
        self.check_wechat_permission(request, "sync")

        def action():
            materials = request.app.sync_materials()
            msg = _("%(count)d materials successfully synchronized")
            return msg % dict(count=len(materials))

        return self._clientaction(
            request, action, _("Sync materials failed with %(exc)s"))

    def has_add_permission(self, request):
        return False

    def get_model_perms(self, request):
        if not request.app.abilities.material:
            return {}
        return super(MaterialAdmin, self).get_model_perms(request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django import forms
from django.contrib import admin
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
import object_tool

from ...models import MessageHandler, MsgLogFlag, Reply, Rule
from ...utils.model import enum2choices
from ..utils import list_property
from ..base import DynamicChoiceForm, WeChatModelAdmin


class MessageHandlerForm(forms.ModelForm):
    log_message = forms.BooleanField(
        label=_("log messages"), initial=False, required=False)

    class Meta(object):
        model = MessageHandler
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        inst = kwargs.get("instance")
        if inst:
            initial = kwargs.get("initial", {})
            initial["log_message"] = inst.log_message
            kwargs["initial"] = initial
        return super(MessageHandlerForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(MessageHandlerForm, self).clean()
        if cleaned_data.get("log_message"):
            cleaned_data["flags"] = MsgLogFlag.LOG_MESSAGE
        else:
            cleaned_data["flags"] = 0
        return cleaned_data

    def save(self, commit=True):
        self.instance.flags = self.cleaned_data["flags"]
        return super(MessageHandlerForm, self).save(commit)


class RuleInline(admin.StackedInline):
    model = Rule
    extra = 0
    min_num = 1

    class RuleForm(DynamicChoiceForm):
        origin_fields = ("type", "weight")

        msg_type = forms.ChoiceField(label=_("message type"),
            choices=enum2choices(Rule.ReceiveMsgType), required=False)
        event = forms.ChoiceField(label=_("event"),
            choices=enum2choices(MessageHandler.EventType), required=False)
        key = forms.CharField(label=_("event key"), required=False)
        pattern = forms.CharField(label=_("pattern"), required=False)
        # media_id = forms.CharField(label=_("media_id"), required=False)

        class Meta(object):
            model = Rule
            fields = ("type", "weight")

        def allowed_fields(self, type, cleaned_data):
            if type in (Rule.Type.CONTAIN, Rule.Type.REGEX, Rule.Type.EQUAL):
                fields = ("pattern", )
            elif type == Rule.Type.EVENT:
                fields = ("event", )
            elif type == Rule.Type.EVENTKEY:
                fields = ("event", "key")
            elif type == Rule.Type.MSGTYPE:
                fields = ("msg_type", )
            else:
                fields = tuple()
            return fields
    form = RuleForm


class ReplyInline(admin.StackedInline):
    model = Reply
    extra = 0

    class ReplyForm(DynamicChoiceForm):
        origin_fields = ("type",)

        program = forms.CharField(label=_("custom program"), required=False)
        url = forms.URLField(label=_("url"), required=False)
        content = forms.CharField(label=_("content"), widget=forms.Textarea,
            required=False)
        media_id = forms.CharField(label=_("media_id"), required=False)
        title = forms.CharField(label=_("title"), required=False)
        description = forms.CharField(label=_("description"),
            widget=forms.Textarea, required=False)
        music_url = forms.URLField(label=_("url"), required=False)
        hq_music_url = forms.URLField(label=_("HQ music url"), required=False)
        thumb_media_id = forms.CharField(label=_("thumb media_id"),
            required=False)

        class Meta(object):
            model = Reply
            fields = ("type", )

        def allowed_fields(self, type, cleaned_data):
            if type == Reply.MsgType.FORWARD:
                fields = ("url", )
            elif type == Reply.MsgType.CUSTOM:
                fields = ("program", )
            elif type == Reply.MsgType.NEWS:
                fields = ("media_id", )
            elif type in (Reply.MsgType.VOICE, Reply.MsgType.IMAGE):
                fields = ("media_id", )
            elif type == Reply.MsgType.VIDEO:
                fields = ("media_id", "title", "description")
            elif type == Reply.MsgType.MUSIC:
                fields = ("title", "description", "music_url", "hq_music_url",
                    "thumb_media_id")
            elif type == Reply.MsgType.TEXT:
                fields = ("content", )
            return fields

        # TODO: 表单验证
    form = ReplyForm


class MessageHandlerAdmin(WeChatModelAdmin):
    __category__ = "messagehandler"
    __model__ = MessageHandler

    class AvailableFilter(admin.SimpleListFilter):
        title = _("available")
        parameter_name = "available"

        def lookups(self, request, model_admin):
            return [(True, _("available"))]

        def queryset(self, request, queryset):
            if self.value():
                now = timezone.now()
                queryset = (queryset.filter(enabled=True)
                    .exclude(starts__gt=now).exclude(ends__lte=now))
            return queryset

    change_form_template = "admin/wechat_django/messagehandler/change_form.html"

    changelist_object_tools = ("sync", )
    list_display = (
        "name", "is_sync",
        list_property(
            "available", boolean=True, short_description=_("available")),
        "enabled", "weight", "starts", "ends", "updated_at", "created_at")
    list_editable = ("weight",)
    list_filter = (AvailableFilter, )
    search_fields = ("name", "rules__content", "replies__content")

    inlines = (RuleInline, ReplyInline)
    fields = ("name", "strategy", "starts", "ends", "enabled", "log_message",
        "weight", "created_at", "updated_at")

    @object_tool.confirm(short_description=_("Sync message handlers"))
    def sync(self, request, queryset):
        self.check_wechat_permission(request, "sync")
        def action():
            handlers = request.app.sync_message_handlers()
            msg = _("%(count)d handlers successfully synchronized")
            return msg % dict(count=len(handlers))

        return self._clientaction(
            request, action, _("Sync handlers failed with %(exc)s"))

    def is_sync(self, obj):
        return obj.src in (MessageHandler.Source.MP, MessageHandler.Source.MENU)
    is_sync.boolean = True
    is_sync.short_description = _("synchronized from wechat")

    def get_fields(self, request, obj=None):
        fields = list(super(MessageHandlerAdmin, self).get_fields(request, obj))
        if not obj:
            fields.remove("created_at")
            fields.remove("updated_at")
        return fields

    def get_readonly_fields(self, request, obj=None):
        rv = super(MessageHandlerAdmin, self).get_readonly_fields(request, obj)
        if obj:
            rv = rv + ("created_at", "updated_at")
        return rv

    def save_model(self, request, obj, form, change):
        obj.src = MessageHandler.Source.SELF
        return super(MessageHandlerAdmin, self).save_model(
            request, obj, form, change)

    form = MessageHandlerForm

    def get_model_perms(self, request):
        if not request.app.abilities.interactable:
            return {}
        return super(MessageHandlerAdmin, self).get_model_perms(request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from wechat_django.admin.base import WeChatModelAdmin


class WeChatPayModelAdmin(WeChatModelAdmin):
    list_filter = ("pay",)

    def get_queryset(self, request):
        return (super(WeChatModelAdmin, self)
            .get_queryset(request).prefetch_related("pay__app")
            .filter(pay__app_id=request.app_id))

    def get_model_perms(self, request):
        if not request.app.abilities.pay:
            return {}
        return super(WeChatPayModelAdmin, self).get_model_perms(request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import os
try:
    from unittest import mock
except ImportError:
    import mock

from django.test import RequestFactory, TestCase

from ..models import WeChatApp


class WeChatTestCaseBase(TestCase):
    def assertCallArgsEqual(self, func, args=(), kwargs=None):
        kwargs = kwargs or {}
        call_args = func.call_args[0]
        call_kwargs = func.call_args[1]
        self.assertEqual(call_args, args)
        self.assertEqual(
            {k: v for k, v in call_kwargs.items() if k in kwargs}, kwargs)

    def load_data(self, path):
        res_file = os.path.join(
            os.path.dirname(__file__), "data", '%s.json' % path)
        with open(res_file, "rb") as f:
            return json.loads(f.read().decode("utf-8"))

    def list_fields(self, model):
        return [f.name for f in model._meta.fields]

    @property
    def base_url(self):
        return "http://localhost/"

    def rf(self, **defaults):
        return RequestFactory(**defaults)


class WeChatTestCase(WeChatTestCaseBase):
    @classmethod
    def setUpTestData(cls):
        super(WeChatTestCase, cls).setUpTestData()
        WeChatApp.objects.create(title="test", name="test",
            appid="appid", appsecret="secret", token="token")
        WeChatApp.objects.create(title="test1", name="test1",
            appid="appid1", appsecret="secret", token="token")
        WeChatApp.objects.create(title="miniprogram", name="miniprogram",
            appid="miniprogram", appsecret="secret",
            type=WeChatApp.Type.MINIPROGRAM)

    def setUp(self):
        self.app = WeChatApp.objects.get_by_name("test")
        self.another_app = WeChatApp.objects.get_by_name("test1")
        self.miniprogram = WeChatApp.objects.get_by_name("miniprogram")

    #region utils
    def _create_handler(self, rules=None, name="", replies=None, app=None,
        **kwargs):
        """:rtype: MessageHandler"""
        from ..models import MessageHandler, Reply, Rule

        if not rules:
            rules = [dict(type=Rule.Type.ALL)]
        if isinstance(rules, dict):
            rules = [rules]
        if isinstance(replies, dict):
            replies = [replies]

        replies = replies or []
        rules = [
            Rule(**rule)
            for rule in rules
        ]
        replies = [
            Reply(**reply)
            for reply in replies
        ]

        app = app or self.app
        return MessageHandler.objects.create_handler(
            app=app,
            name=name,
            rules=rules,
            replies=replies,
            **kwargs
        )

    def _msg2info(self, message, app=None, **kwargs):
        """:rtype: WeChatMessageInfo"""
        from ..models import WeChatMessageInfo
        return WeChatMessageInfo(
            _app=app or self.app,
            _message=message,
            **{
                "_" + k: v
                for k, v in kwargs.items()
            }
        )
    #endregion



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

import django
from django import forms
from django.contrib import admin, messages
from django.contrib.admin.actions import delete_selected
from django.contrib.admin.views.main import ChangeList
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from object_tool import CustomObjectToolModelAdmin
import six
from wechatpy.exceptions import WeChatClientException

from ..models.permission import get_user_permissions


registered_admins = []


class RecursiveDeleteActionMixin(object):
    """逐一删除混合类"""
    def get_actions(self, request):
        actions = super(RecursiveDeleteActionMixin, self).get_actions(request)
        if "delete_selected" in actions:
            actions["delete_selected"] = (
                RecursiveDeleteActionMixin.delete_selected_recusively,
                actions["delete_selected"][1],
                actions["delete_selected"][2]
            )
        return actions

    def delete_selected_recusively(self, request, queryset):
        """逐一删除"""
        if not request.POST.get("post"):
            resp = delete_selected(self, request, queryset)
            resp.context_data.update(
                wechat_app=request.app,
                wechat_app_id=request.app_id
            )
            return resp

        with transaction.atomic():
            for o in queryset.all():
                try:
                    if not self.has_delete_permission(request, o):
                        raise PermissionDenied
                    o.delete()
                except WeChatClientException:
                    msg = _("delete %(category)s failed: %(obj)s") % dict(
                        category=self.model.verbose_name_plural,
                        obj=o
                    )
                    request.app.logger("admin").warning(msg, exc_info=True)
                    raise
    delete_selected.short_description = _("delete selected")


def has_wechat_permission(request, app, category="", operate="", obj=None):
    """
    检查用户是否具有某一微信权限
    :type request: django.http.request.HttpRequest
    """
    if request.user.is_superuser:
        return True
    perms = get_user_permissions(request.user, app)
    needs = {category, "{0}_{1}".format(category, operate)}
    return bool(needs.intersection(perms))


class WeChatChangeList(ChangeList):
    def __init__(self, request, *args, **kwargs):
        super(WeChatChangeList, self).__init__(request, *args, **kwargs)
        self.request = request

    def url_for_result(self, result):
        view = "admin:%s_%s_change" % (
            self.opts.app_label, self.opts.model_name)
        kwargs = dict(
            object_id=getattr(result, self.pk_attname),
            wechat_app_id=self.request.app_id
        )
        return reverse(
            view, kwargs=kwargs, current_app=self.model_admin.admin_site.name)


class WeChatModelAdminMetaClass(forms.MediaDefiningClass):
    def __new__(cls, name, bases, attrs):
        model = attrs.pop("__model__", None)
        self = super(WeChatModelAdminMetaClass, cls).__new__(
            cls, name, bases, attrs)
        if name != "WeChatModelAdmin" and model:
            registered_admins.append((model, self))
        return self


class WeChatModelAdmin(six.with_metaclass(WeChatModelAdminMetaClass, CustomObjectToolModelAdmin)):
    """所有微信相关业务admin的基类

    并且通过request.app_id及request.app拿到app信息
    """
    change_form_template = "admin/wechat_django/change_form.html"
    change_list_template = "admin/wechat_django/change_list.html"
    objecttool_form_template = "admin/wechat_django/objecttool_form.html"

    #region view
    def get_changelist(self, request):
        return WeChatChangeList

    def get_urls(self):
        urlpatterns = super(WeChatModelAdmin, self).get_urls()
        # django 1.11 替换urlpattern为命名式的
        if django.VERSION[0] < 2:
            for pattern in urlpatterns:
                pattern._regex = pattern._regex.replace(
                    "(.+)", "(?P<object_id>.+)")
        return urlpatterns

    def _clientaction(self, request, action, failed_msg, kwargs=None):
        """调用微信请求action"""
        kwargs = kwargs or dict()
        try:
            msg = action()
            self.message_user(request, msg)
        except Exception as e:
            kwargs.update(exc=e)
            msg = failed_msg % kwargs
            if isinstance(e, WeChatClientException):
                request.app.logger("admin").warning(msg, exc_info=True)
            else:
                request.app.logger("admin").error(msg, exc_info=True)
            self.message_user(request, msg, level=messages.ERROR)
    #endregion

    #region model
    def get_queryset(self, request):
        return (super(WeChatModelAdmin, self)
            .get_queryset(request).filter(app_id=request.app_id))

    def save_model(self, request, obj, form, change):
        if not change:
            obj.app_id = request.app_id
        return super(WeChatModelAdmin, self).save_model(
            request, obj, form, change)
    #endregion

    #region permissions
    def check_wechat_permission(self, request, operate="", category="", obj=None):
        if not self.has_wechat_permission(request, operate, category, obj):
            raise PermissionDenied

    def has_wechat_permission(self, request, operate="", category="", obj=None):
        if not hasattr(request, "app"):
            return False
        app = request.app
        category = category or self.__category__
        return has_wechat_permission(request, app, category, operate, obj)

    def has_add_permission(self, request):
        return self.has_wechat_permission(request, "add")

    def has_change_permission(self, request, obj=None):
        return self.has_wechat_permission(request, "change", obj=obj)

    def has_delete_permission(self, request, obj=None):
        return self.has_wechat_permission(request, "delete", obj=obj)

    def has_module_permission(self, request):
        """是否拥有任意本公众号管理权限"""
        if not hasattr(request, "app"):
            return False
        return bool(get_user_permissions(request.user, request.app))
    #endregion


class DynamicChoiceForm(forms.ModelForm):
    content_field = "_content"
    type_field = "type"
    origin_fields = tuple()

    def __init__(self, *args, **kwargs):
        inst = kwargs.get("instance")
        if inst:
            initial = kwargs.get("initial", {})
            initial.update(getattr(inst, self.content_field))
            kwargs["initial"] = initial
        super(DynamicChoiceForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(DynamicChoiceForm, self).clean()
        if self.type_field not in cleaned_data:
            self.add_error(self.type_field, "")
            return
        type = cleaned_data[self.type_field]
        fields = self.allowed_fields(type, cleaned_data)

        content = dict()
        for k in set(cleaned_data.keys()).difference(self.origin_fields):
            if k in fields:
                content[k] = cleaned_data[k]
                del cleaned_data[k]
        cleaned_data[self.content_field] = content
        return cleaned_data

    def allowed_fields(self, type, cleaned_data):
        raise NotImplementedError()

    def save(self, commit=True, *args, **kwargs):
        model = super(DynamicChoiceForm, self).save(False, *args, **kwargs)
        setattr(model, self.content_field,
            self.cleaned_data[self.content_field])
        if commit:
            model.save()
        return model



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4

import xmltodict

from wechat_django.sites.wechat import default_site
from ..exceptions import WeChatPayNotifyError
from ..models import UnifiedOrderResult
from ..notify import NotifyViewSet
from .base import mock, WeChatPayTestCase


class NotifyTestCase(WeChatPayTestCase):
    sign = "sign"

    def test_response(self):
        """测试响应"""
        url = self.app.build_url(
            "order_notify", kwargs=dict(payname=self.app.pay.name))
        # 测试成功响应
        order = self.app.pay.create_order(**self.minimal_example)
        xml = self.success_order_notify(self.app.pay, order)
        with mock.patch("wechatpy.pay.calculate_signature") as m:
            m.return_value = self.sign
            resp = self.client.post(
                url, data=xml, content_type="text/xml")
            data = xmltodict.parse(resp.content)["xml"]
            self.assertEqual(data["return_code"], "SUCCESS")
            self.assertNotEqual(data["return_msg"], "OK")

        # 测试失败响应
        resp = self.client.get(url)
        data = xmltodict.parse(resp.content)["xml"]
        self.assertEqual(data["return_code"], "FAIL")
        self.assertNotEqual(data["return_msg"], "OK")

    def test_prepare_request(self):
        """测试请求预处理"""
        appname = self.app.name
        url = self.app.build_url(
            "order_notify", kwargs=dict(payname=self.app.pay.name))
        # 测试签名正确
        order = self.app.pay.create_order(**self.minimal_example)
        xml = self.success_order_notify(self.app.pay, order)
        request = self.rf().post(
            url, data=xml, content_type="text/xml")

        viewset = NotifyViewSet(default_site)
        with mock.patch("wechatpy.pay.calculate_signature") as m:
            m.return_value = self.sign
            pay, data = viewset._prepare(
                request, self.app.name, self.app.pay.name)

        # 测试签名错误
        self.assertRaises(
            WeChatPayNotifyError, viewset._prepare,
            request, self.app.name, self.app.pay.name)

    def test_notify_order(self):
        """测试订单回调通知"""
        url = self.app.build_url(
            "order_notify", kwargs=dict(payname=self.app.pay.name))
        order = self.app.pay.create_order(**self.minimal_example)
        xml = self.success_order_notify(self.app.pay, order)
        request = self.rf().post(
            url, data=xml, content_type="text/xml")

        viewset = NotifyViewSet(default_site)
        with mock.patch("wechatpy.pay.calculate_signature") as m:
            m.return_value = self.sign
            pay, data = viewset._prepare(
                request, self.app.name, self.app.pay.name)

        viewset.order_notify(request, pay, data)
        self.assertEqual(order.result.transaction_id, order.out_trade_no)
        self.assertEqual(
            order.result.trade_state, UnifiedOrderResult.State.SUCCESS)

    @property
    def minimal_example(self):
        return dict(
            body="body",
            out_trade_no=str(uuid4()),
            total_fee=101
        )

    def success_order_notify(self, pay, order):
        return """
            <xml>
                <appid><![CDATA[{appid}]]></appid>
                <bank_type><![CDATA[CFT]]></bank_type>
                <cash_fee><![CDATA[1]]></cash_fee>
                <fee_type><![CDATA[CNY]]></fee_type>
                <is_subscribe><![CDATA[N]]></is_subscribe>
                <mch_id><![CDATA[{mch_id}]]></mch_id>
                <nonce_str><![CDATA[nonce_str]]></nonce_str>
                <openid><![CDATA[{openid}]]></openid>
                <out_trade_no><![CDATA[{out_trade_no}]]></out_trade_no>
                <result_code><![CDATA[SUCCESS]]></result_code>
                <return_code><![CDATA[SUCCESS]]></return_code>
                <sign><![CDATA[{sign}]]></sign>
                <time_end><![CDATA[20190618223614]]></time_end>
                <total_fee>101</total_fee>
                <trade_type><![CDATA[JSAPI]]></trade_type>
                <transaction_id><![CDATA[{out_trade_no}]]></transaction_id>
            </xml>
        """.format(
            appid=pay.appid,
            mch_id=pay.mch_id,
            openid="openid",
            out_trade_no=order.out_trade_no,
            sign=self.sign
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from wechatpy.events import BaseEvent

from ..utils.model import enum2choices
from . import Rule, WeChatApp, WeChatModel, WeChatUser


class MessageLog(WeChatModel):
    class Direct(object):
        USER2APP = False
        APP2USER = True

    app = m.ForeignKey(WeChatApp, on_delete=m.CASCADE)
    user = m.ForeignKey(WeChatUser, on_delete=m.CASCADE)

    msg_id = m.BigIntegerField(_("msgid"), null=True)
    type = m.CharField(
        _("message type"), max_length=24,
        choices=enum2choices(Rule.ReceiveMsgType))
    content = JSONField(_("content"))
    direct = m.BooleanField(_("direct"), default=Direct.USER2APP)

    raw = m.TextField(null=True, blank=True, default=None)

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)

    class Meta(object):
        verbose_name = _("message log")
        verbose_name_plural = _("message logs")

        index_together = (("app", "created_at"),)
        ordering = ("app", "-created_at")

    @classmethod
    def from_message_info(cls, message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        return cls._from_message(
            message_info.message,
            message_info.app,
            message_info.user,
            # message_info.raw
        )

    @classmethod
    def _from_message(cls, message, app, user, raw=""):
        content = {
            key: getattr(message, key)
            for key in message._fields
            if key not in ("id", "source", "target", "create_time", "time")
        }

        if isinstance(message, BaseEvent):
            content["event"] = message.event

        kwargs = dict(
            app=app,
            user=user,
            msg_id=message.id,
            type=message.type,
            content=content,
            raw=raw,
            direct=cls.Direct.USER2APP
        )
        if message.time:
            kwargs["created_at"] = timezone.datetime.fromtimestamp(
                message.time)
        return cls.objects.create(**kwargs)

    @classmethod
    def from_reply(cls, reply, app, user):
        """:type reply: wechatpy.replies.BaseReply"""
        content = {
            key: getattr(reply, key)
            for key in reply._fields
            if key not in ("id", "source", "target", "create_time", "time")
        }

        kwargs = dict(
            app=app,
            user=user,
            type=reply.type,
            content=content,
            # raw=reply.render(),
            direct=cls.Direct.APP2USER
        )
        if reply.time:
            kwargs["created_at"] = timezone.datetime.fromtimestamp(
                reply.time)
        return cls.objects.create(**kwargs)

    def __str__(self):
        return _("%(type)s消息: %(msg_id)s") % dict(
            type=self.type,
            msg_id=self.msg_id
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from contextlib import contextmanager
import os
from tempfile import NamedTemporaryFile

from wechatpy import WeChatPay as _Pay
from wechatpy.exceptions import WeChatPayException


@contextmanager
def load_cert(self):
    if self.pay.mch_cert and self.pay.mch_key:
        # 证书文件要在硬盘上读取
        # 自动delete在windows下读取时报PermissionDenied
        # TODO: 可以考虑第一次使用时写入证书 析构时移除证书 减少IO
        with NamedTemporaryFile("wb", delete=False) as mch_cert,\
             NamedTemporaryFile("wb", delete=False) as mch_key:
            mch_cert.write(self.pay.mch_cert)
            mch_cert.flush()
            mch_cert.close()
            mch_key.write(self.pay.mch_key)
            mch_key.flush()
            mch_key.close()
            self.mch_cert = mch_cert.name
            self.mch_key = mch_key.name
            try:
                yield
            finally:
                self.mch_key and os.path.exists(self.mch_key)\
                    and os.remove(self.mch_key)
                self.mch_key = None
                self.mch_cert and os.path.exists(self.mch_cert)\
                    and os.remove(self.mch_cert)
                self.mch_cert = None
    else:
        yield


class WeChatPayClient(_Pay):
    def __init__(self, pay):
        """:type pay: wechat_django.models.WeChatPay"""
        self.pay = pay
        kwargs = dict(
            appid=pay.appid,
            api_key=pay.api_key,
            mch_id=pay.mch_id,
            sub_mch_id=pay.sub_mch_id
        )
        if pay.mch_app_id:
            kwargs["sub_appid"] = pay.sub_appid

        super(WeChatPayClient, self).__init__(**kwargs)

    def _request(self, method, url_or_endpoint, **kwargs):
        logger = self.pay.app.logger("client")
        log_args = dict(data=kwargs.get("data", ""))
        with load_cert(self):
            try:
                rv = super(WeChatPayClient, self)._request(
                    method, url_or_endpoint, **kwargs)
            except WeChatPayException as e:
                log_args["err"] = e
                logger.warning(
                    "An error occurred when send wechat pay"
                    "request: %s" % log_args, exc_info=True)
                raise
            except Exception as e:
                log_args["err"] = e
                logger.error(
                    "An unexcept error occurred when send wechat pay"
                    "request: %s" % log_args, exc_info=True)
                raise
            else:
                log_args["result"] = rv
                logger.debug("A WeChat pay request sent: %s" % log_args)
            return rv


class WeChatPaySandboxClient(WeChatPayClient):
    def __init__(self, pay):
        super(WeChatPaySandboxClient, self).__init__(pay)
        self.sandbox = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _


permissions = {
    "pay_full": _("Can full control %(appname)s WeChat pay"),
    "pay_manage": _("Can manage %(appname)s WeChat pay"),
    "pay_order": _("Can view %(appname)s WeChat pay orders")
}


permission_required = {
    "pay_full": {
        "pay_manage",
        "pay_order"
    },
    "pay_manage": {
        "wechat_django_pay.add_wechatpay",
        "wechat_django_pay.delete_wechatpay",
        "wechat_django_pay.change_wechatpay"
    }
}



# -*- coding: utf-8 -*-
# Generated by Django 2.1.7 on 2019-05-24 03:49
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import wechat_django.pay.models.base

from wechat_django.models.permission import upgrade_perms, downgrade_perms


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wechat_django', '0004_templatemessage'),
    ]

    operations = [
        migrations.CreateModel(
            name='UnifiedOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('device_info', models.CharField(max_length=32, null=True, verbose_name='device_info')),
                ('body', models.CharField(max_length=128, verbose_name='order body')),
                ('detail', models.TextField(max_length=6000, null=True, verbose_name='order detail')),
                ('out_trade_no', models.CharField(max_length=32, verbose_name='out_trade_no')),
                ('fee_type', models.CharField(default='CNY', max_length=16, null=True, verbose_name='fee_type')),
                ('total_fee', models.PositiveIntegerField(verbose_name='total_fee')),
                ('spbill_create_ip', models.CharField(max_length=64, verbose_name='spbill_create_ip')),
                ('time_start', wechat_django.pay.models.base.PayDateTimeField(default=django.utils.timezone.now, verbose_name='time_start')),
                ('time_expire', wechat_django.pay.models.base.PayDateTimeField(verbose_name='time_expire')),
                ('goods_tag', models.CharField(max_length=32, null=True, verbose_name='goods_tag')),
                ('trade_type', models.CharField(choices=[('APP', 'APP'), ('JSAPI', 'JSAPI'), ('MICROPAY', 'MICROPAY'), ('MWEB', 'MWEB'), ('NATIVE', 'NATIVE')], max_length=16, verbose_name='trade_type')),
                ('product_id', models.CharField(max_length=32, null=True, verbose_name='product id')),
                ('limit_pay', models.CharField(choices=[('no_credit', 'NOCREDIT'), (None, 'NONE')], default=None, max_length=32, null=True, verbose_name='limit_pay')),
                ('openid', models.CharField(max_length=128, null=True, verbose_name='openid')),
                ('sub_openid', models.CharField(max_length=128, null=True, verbose_name='sub openid')),
                ('receipt', models.CharField(choices=[(None, 'NONE'), ('Y', 'TRUE')], default=None, max_length=8, null=True, verbose_name='recept')),
                ('scene_info', jsonfield.fields.JSONField(max_length=256, null=True, verbose_name='scene_info')),
                ('comment', models.TextField(blank=True, verbose_name='comment')),
                ('ext_info', jsonfield.fields.JSONField(default=dict, editable=False)),
                ('_call_args', jsonfield.fields.JSONField(db_column='call_args', default=dict, editable=False, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated at')),
            ],
            options={
                'verbose_name': 'Unified order',
                'verbose_name_plural': 'Unified orders',
            },
        ),
        migrations.CreateModel(
            name='UnifiedOrderResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('transaction_id', models.CharField(max_length=32, verbose_name='transaction_id', null=True)),
                ('trade_state', models.CharField(choices=[('CLOSED', 'CLOSED'), ('FAIL', 'FAIL'), ('NOTPAY', 'NOTPAY'), ('PAYERROR', 'PAYERROR'), ('REFUND', 'REFUND'), ('REVOKED', 'REVOKED'), ('SUCCESS', 'SUCCESS'), ('USERPAYING', 'USERPAYING')], max_length=32, verbose_name='trade_state')),
                ('time_end', wechat_django.pay.models.base.PayDateTimeField(null=True, verbose_name='pay time_end')),
                ('settlement_total_fee', models.PositiveIntegerField(null=True, verbose_name='settlement_total_fee')),
                ('cash_fee', models.PositiveIntegerField(verbose_name='cash_fee', null=True)),
                ('cash_fee_type', models.CharField(max_length=16, null=True, verbose_name='cash_fee_type')),
                ('coupon_fee', models.PositiveIntegerField(null=True, verbose_name='coupon_fee')),
                ('bank_type', models.CharField(max_length=16, verbose_name='bank_type', null=True)),
                ('detail', models.TextField(max_length=8192, null=True, verbose_name='detail')),
                ('is_subscribe', wechat_django.pay.models.base.PayBooleanField(null=True, verbose_name='is_subscribe')),
                ('sub_is_subscribe', wechat_django.pay.models.base.PayBooleanField(null=True, verbose_name='sub_is_subscribe')),
                ('ext_info', jsonfield.fields.JSONField(default=dict, editable=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated at')),
                ('order', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='result', to='wechat_django_pay.UnifiedOrder')),
            ],
            options={
                'verbose_name': 'Unified order result',
                'verbose_name_plural': 'Unified order results',
            },
        ),
        migrations.CreateModel(
            name='WeChatPay',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(blank=True, help_text='商户号标识,用于后台辨识商户号', max_length=16, verbose_name='title')),
                ('name', models.CharField(default='default', help_text='商户号程序标识', max_length=16, verbose_name='name')),
                ('weight', models.IntegerField(default=0, verbose_name='weight')),
                ('mch_id', models.CharField(help_text='微信支付分配的商户号', max_length=32, verbose_name='mch_id')),
                ('api_key', models.CharField(help_text='商户号key', max_length=128, verbose_name='WeChatPay api_key')),
                ('sub_mch_id', models.CharField(blank=True, help_text='子商户号，受理模式下填写', max_length=32, null=True, verbose_name='sub_mch_id')),
                ('mch_app_id', models.CharField(blank=True, help_text='微信分配的主商户号appid，受理模式下填写', max_length=32, null=True, verbose_name='mch_app_id')),
                ('mch_cert', models.BinaryField(blank=True, null=True, verbose_name='mch_cert')),
                ('mch_key', models.BinaryField(blank=True, null=True, verbose_name='mch_key')),
                ('app', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pays', to='wechat_django.WeChatApp')),
            ],
            options={
                'verbose_name': 'WeChat pay',
                'verbose_name_plural': 'WeChat pay',
                'ordering': ('app', '-weight', 'pk'),
            },
        ),
        migrations.AddField(
            model_name='unifiedorder',
            name='pay',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='orders', to='wechat_django_pay.WeChatPay'),
        ),
        migrations.AlterUniqueTogether(
            name='wechatpay',
            unique_together={('app', 'name')},
        ),
        migrations.AlterUniqueTogether(
            name='unifiedorder',
            unique_together={('pay', 'out_trade_no')},
        ),
        migrations.AlterIndexTogether(
            name='unifiedorder',
            index_together={('pay', 'created_at')},
        ),
    ]

    migrations.RunPython(
        lambda *args, **kwargs: upgrade_perms("pay_full", "pay_manage", "pay_order"),
        lambda *args, **kwargs: downgrade_perms("pay_full", "pay_manage", "pay_order"),
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from httmock import all_requests, HTTMock, response, urlmatch

TESTFORBIDDEN = -99999


def common_interceptor(callback, **kwargs):
    decorator = urlmatch(**kwargs) if kwargs else all_requests
    mock = decorator(callback)
    return HTTMock(mock)


def wechatapi(api, data="", callback=None):
    @urlmatch(netloc=r"(.*\.)?api\.weixin\.qq\.com$", path=api)
    def wechatapi_mock(url, request):
        if url.path != api:
            return response(404)
        headers = {
            "Content-Type": "application/json"
        }
        resp = response(200, data, headers)
        if callback:
            callback(url, request, response)
        return resp

    return HTTMock(wechatapi_mock)


def wechatapi_accesstoken(callback=None):
    return wechatapi("/cgi-bin/token", {
        "access_token": "ACCESS_TOKEN",
        "expires_in": 7200
    }, callback)


def wechatapi_error(api):
    return wechatapi(api, {
        "errcode": TESTFORBIDDEN,
        "errmsg": "",
    })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..admin.base import WeChatModelAdmin
from ..models import WeChatUser
from ..sites.admin import WeChatAdminSite
from .base import WeChatTestCase


class AdminBaseTestCase(WeChatTestCase):
    def test_get_request_params(self):
        """测试从admin的request请求中获取参数的方法"""
        pass

    def test_get_queryset(self):
        """测试queryset拿到的是本app的queryset"""
        openid = "openid"
        WeChatUser.objects.create(app=self.app, openid=openid)
        WeChatUser.objects.create(app=self.another_app, openid=openid)

        site = WeChatAdminSite()
        model_admin = WeChatModelAdmin(WeChatUser, site)
        request = self.rf().get("/")
        request.app_id = self.app.id
        request.app = self.app
        queryset = model_admin.get_queryset(request)
        self.assertEqual(
            queryset.filter(openid=openid).count(),
            1
        )
        self.assertEqual(
            queryset.count(),
            queryset.filter(app=self.app).count()
        )
        self.assertEqual(
            queryset.exclude(app=self.app).count(),
            0
        )

    def test_add_model(self):
        """测试新增数据时,会把appid附上"""
        pass



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import unicode_literals

from .constants import MsgLogFlag, MsgType

from .app import WeChatApp
from .permission import permissions
from .base import WeChatModel, appmethod
from .template import Template
from .user import WeChatUser
from .usertag import UserTag
from .request import WeChatInfo, WeChatMessageInfo, WeChatOAuthInfo
from .material import Material
from .article import Article
from .messagehandler import MessageHandler
from .reply import Reply
from .rule import Rule
from .messagelog import MessageLog
from .menu import Menu
from .session import Session



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

ALLOWED_HOSTS = ["example.com"]

DEBUG = True

SECRET_KEY = "fake-key"


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "wechat_django",
    "wechat_django.pay"
]


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

TIME_ZONE = 'Asia/Shanghai'

USE_TZ = True

STATIC_URL = '/static/'

ROOT_URLCONF = "wechat_django.tests.urls"

logging.disable(logging.CRITICAL)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


def next_chunk(iterator, count=100):
    rv = []
    for item in iterator:
        rv.append(item)
        if len(rv) >= count:
            yield rv
            rv = []
    if rv:
        yield rv


class Static:
    __caches = dict()

    def __new__(cls, obj):
        if obj not in cls.__caches:
            cls.__caches[obj] = obj
        return cls.__caches[obj]

    def __init__(self, obj):
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-03-09 16:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

from ..models.permission import upgrade_perms, downgrade_perms


class Migration(migrations.Migration):

    dependencies = [
        ('wechat_django', '0001_init'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserTag',
            fields=[
                ('_id', models.AutoField(primary_key=True, serialize=False)),
                ('id', models.IntegerField(verbose_name='tag id')),
                ('name', models.CharField(max_length=30, verbose_name='tag name')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
            ],
        ),
        migrations.AlterField(
            model_name='wechatapp',
            name='appsecret',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='AppSecret'),
        ),
        migrations.AddField(
            model_name='usertag',
            name='app',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='user_tags', to='wechat_django.WeChatApp'),
        ),
        migrations.AddField(
            model_name='usertag',
            name='users',
            field=models.ManyToManyField(related_name='tags', to='wechat_django.WeChatUser'),
        ),
        migrations.AlterUniqueTogether(
            name='usertag',
            unique_together=set([('app', 'name'), ('app', 'id')]),
        ),
    ]

    # 追加两个权限
    migrations.RunPython(
        lambda *args, **kwargs: upgrade_perms("usertag", "usertag_sync"),
        lambda *args, **kwargs: downgrade_perms("usertag", "usertag_sync"),
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.http import response
from django.urls import reverse
from django.urls.resolvers import get_ns_resolver

from .base import WeChatTestCase


class UtilDecoratorTestCase(WeChatTestCase):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import transaction
from wechatpy.client.api import WeChatGroup, WeChatTag

from ..models import UserTag, WeChatUser
from .base import mock, WeChatTestCase


class UserTestCase(WeChatTestCase):
    def test_sync(self):
        """测试同步用户标签"""
        def assertTagSyncSuccess(custom_tags):
            with mock.patch.object(WeChatGroup, "get"),\
                mock.patch.object(WeChatTag, "update"),\
                mock.patch.object(WeChatTag, "create"):
                remote_tags = self.base_tags[:] + [
                    dict(
                        id=id,
                        name=name,
                        count=0
                    )
                    for id, name in custom_tags.items()
                ]
                WeChatGroup.get.return_value = remote_tags
                # 不会调用远程接口
                WeChatTag.update.side_effect = Exception()
                WeChatTag.create.side_effect = Exception()

                UserTag.sync(self.app)
                tags = UserTag.objects.filter(app=self.app).all()
                self.assertEqual(len(tags), len(remote_tags))
                # 检查每个tag被正确插入
                for tag in remote_tags:
                    UserTag.objects.get(app=self.app, id=tag["id"], name=tag["name"])

        assertTagSyncSuccess({101: "group1", 102: "group2"})
        assertTagSyncSuccess({103: "group1", 102: "group2"})
        assertTagSyncSuccess({104: "group3"})

    def test_sync_users(self):
        """测试同步标签下的用户"""
        with mock.patch.object(WeChatTag, "iter_tag_users"),\
            mock.patch.object(WeChatUser, "upsert_users"):

            WeChatTag.iter_tag_users.return_value = ["openid1", "openid2"]
            WeChatUser.upsert_users.return_value = ["openid1", "openid2"]
            tag = UserTag.objects.create(
                app=self.app, name="tag", id=101, _tag_local=True)
            del tag._tag_local
            self.assertEqual(tag.sync_users(False), ["openid1", "openid2"])

    def test_change_user_tags(self):
        """测试用户标签变更"""
        tag_user_err = "tag_user_err"
        untag_user_err = "untag_user_err"

        user = WeChatUser.objects.create(app=self.app, openid="openid1")
        tag1 = UserTag.objects.create(
            app=self.app, id=101, name="tag1", _tag_local=True)
        del tag1._tag_local
        tag2 = UserTag.objects.create(
            app=self.app, id=102, name="tag2", _tag_local=True)
        del tag2._tag_local

        with mock.patch.object(WeChatTag, "tag_user"),\
            mock.patch.object(WeChatTag, "untag_user"):

            # 测试用户添加标签
            WeChatTag.untag_user.side_effect = Exception(untag_user_err)
            user.tags.set((tag1,))
            user.tags.filter(id=tag1.id).get()
            self.assertEqual(WeChatTag.tag_user.call_args, ((tag1.id, user.openid),))

            user.tags.set((tag1, tag2))
            self.assertEqual(user.tags.count(), 2)
            self.assertEqual(
                WeChatTag.tag_user.call_args, ((tag2.id, user.openid),))

            # 测试用户加减标签
            WeChatTag.untag_user.side_effect = None
            WeChatTag.tag_user.side_effect = Exception(tag_user_err)
            user.tags.set((tag2, ))
            self.assertEqual(user.tags.count(), 1)
            self.assertEqual(
                WeChatTag.untag_user.call_args, ((tag1.id, user.openid),))

            # 测试用户加减标签异常
            WeChatTag.untag_user.side_effect = Exception(untag_user_err)

            with transaction.atomic(), self.assertRaises(Exception) as context:
                user.tags.add(tag1)
            self.assertIn(tag_user_err, str(context.exception))
            self.assertRaises(
                UserTag.DoesNotExist,
                lambda: user.tags.get(app=self.app, id=tag1.id))

            with transaction.atomic(), self.assertRaises(Exception) as context:
                user.tags.remove(tag2)
            self.assertIn(untag_user_err, str(context.exception))
            user.tags.get(app=self.app, id=tag2.id)

    def test_tag_users(self):
        """测试标签用户变更"""
        tag_user_err = "tag_user_err"
        untag_user_err = "untag_user_err"

        user1 = WeChatUser.objects.create(app=self.app, openid="openid1")
        user2 = WeChatUser.objects.create(app=self.app, openid="openid2")
        tag = UserTag.objects.create(
            app=self.app, id=101, name="tag1", _tag_local=True)
        del tag._tag_local

        with mock.patch.object(WeChatTag, "tag_user"),\
            mock.patch.object(WeChatTag, "untag_user"):

            # 测试标签添加用户
            WeChatTag.untag_user.side_effect = Exception(untag_user_err)
            tag.users.set((user1,))
            tag.users.filter(openid=user1.openid).get()
            self.assertEqual(WeChatTag.tag_user.call_args, ((tag.id, [user1.openid]),))

            tag.users.set((user1, user2))
            self.assertEqual(tag.users.count(), 2)
            self.assertEqual(
                WeChatTag.tag_user.call_args, ((tag.id, [user2.openid]),))

            # 测试标签加减用户
            WeChatTag.untag_user.side_effect = None
            WeChatTag.tag_user.side_effect = Exception(tag_user_err)
            tag.users.set((user2, ))
            self.assertEqual(tag.users.count(), 1)
            self.assertEqual(
                WeChatTag.untag_user.call_args, ((tag.id, [user1.openid]),))

            # 测试标签加减用户异常
            WeChatTag.untag_user.side_effect = Exception(untag_user_err)

            with transaction.atomic(), self.assertRaises(Exception) as context:
                tag.users.add(user1)
            self.assertIn(tag_user_err, str(context.exception))
            self.assertRaises(
                WeChatUser.DoesNotExist,
                lambda: tag.users.get(app=self.app, openid=user1.openid))

            with transaction.atomic(), self.assertRaises(Exception) as context:
                tag.users.remove(user2)
            self.assertIn(untag_user_err, str(context.exception))
            tag.users.get(app=self.app, openid=user2.openid)

    def test_edit_tag(self):
        """测试标签的增删改"""
        id = 101
        name = "group1"
        # 新增
        with mock.patch.object(WeChatTag, "create"):
            WeChatTag.create.return_value = dict(
                id=id,
                name=name
            )
            UserTag(app=self.app, name=name).save()
            tag = UserTag.objects.get(app=self.app, name=name)
            self.assertEqual(tag.id, id)
            self.assertEqual(WeChatTag.create.call_args, ((name,),))

        name = "edit"
        with mock.patch.object(WeChatTag, "update"):
            tag.name = name
            tag.save()
            tag = UserTag.objects.get(app=self.app, name=name)
            self.assertEqual(tag.id, id)
            self.assertEqual(WeChatTag.update.call_args, ((id, name),))

        with mock.patch.object(WeChatTag, "delete"):
            tag.delete()
            self.assertRaises(
                UserTag.DoesNotExist,
                lambda: UserTag.objects.get(app=self.app, id=id))
            self.assertEqual(WeChatTag.delete.call_args, ((id,),))

    @property
    def base_tags(self):
        return [
            {
                "id": 0,
                "name": "未分组",
                "count": 0
            },
            {
                "id": 1,
                "name": "黑名单",
                "count": 0
            },
            {
                "id": 2,
                "name": "星标组",
                "count": 0
            }
        ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin
from django.conf.urls import url

import wechat_django

urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r"^wechat/", wechat_django.sites.wechat.urls)
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import timedelta
import time

from django.utils import timezone
from wechatpy import messages

from ..handler import message_handler
from ..models import MessageHandler, Reply, Rule, WeChatMessageInfo
from .base import WeChatTestCase
from .interceptors import (common_interceptor, wechatapi,
    wechatapi_accesstoken, wechatapi_error)


@message_handler
def debug_handler(message):
    return "success"


@message_handler("test")
def app_only_handler(message):
    return "success"


def forbidden_handler(message):
    return ""


class HandlerTestCase(WeChatTestCase):
    def test_available(self):
        """测试handler有效性"""
        rule = dict(type=Rule.Type.ALL)
        now = timezone.now()
        day = timedelta(days=1)
        handler_not_begin = self._create_handler(rule, starts=now + day)
        handler_ended = self._create_handler(rule, ends=now - day)
        handler_disabled = self._create_handler(rule, enabled=False)
        handler_available = self._create_handler(
            rule, starts=now - day, ends=now + day)

        msg_info = self._msg2info(messages.TextMessage("abc"))
        self.assertFalse(handler_not_begin.is_match(msg_info))
        self.assertFalse(handler_ended.is_match(msg_info))
        self.assertFalse(handler_disabled.is_match(msg_info))
        self.assertTrue(handler_available.is_match(msg_info))

        matches = MessageHandler.matches(msg_info)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0], handler_available)

    def test_sync(self):
        """测试同步"""
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4 as uuid

from django.contrib.admin import site
from django.contrib.auth.models import ContentType, Permission, User
from django.urls import ResolverMatch

from wechat_django.models import permission as pm, WeChatApp
from wechat_django.pay import admin as pay_admin
from wechat_django.pay.admin.payapp import WeChatPayInline, WeChatAppWithPayAdmin
from wechat_django.pay import models as pay_models
from .base import WeChatPayTestCase


class PermissionTestCase(WeChatPayTestCase):
    def test_app_menu(self):
        """测试各微信号菜单权限"""
        request = self.rf().get("/admin/wechat_django/%s/" % self.app.id)
        request.app = self.app
        n_user = self._create_user()

        permission_map = dict(
            pay_order=(pay_models.UnifiedOrder, pay_admin.order.OrderAdmin)
        )
        for permission, ma in permission_map.items():
            perm_name = pm.get_perm_name(self.app, permission)
            Model, Admin = ma
            p_user = self._create_user(perm_name)
            request.user = p_user
            admin = Admin(Model, site)
            perms = admin.get_model_perms(request)
            self.assertTrue(perms["change"])
            request.user = n_user
            perms = admin.get_model_perms(request)
            self.assertFalse(perms.get("change"))

    def test_pay_manage(self):
        """测试支付权限"""
        perm_name = pm.get_perm_name(self.app, "pay_manage")
        p_user = self._create_user(perm_name)
        n_user = self._create_user()
        admin = WeChatAppWithPayAdmin(WeChatApp, site)

        request = self.rf().get("/admin/wechat_django/apps?id=" + str(self.app.id))
        request.user = p_user
        inlines = admin.get_inline_instances(request, self.app)
        self.assertTrue(list(filter(
            lambda o: isinstance(o, WeChatPayInline), inlines)))

        request.user = n_user
        inlines = admin.get_inline_instances(request, self.app)
        self.assertFalse(list(filter(
            lambda o: isinstance(o, WeChatPayInline), inlines)))

    def _create_user(self, perm_name=None):
        user = User.objects.create_user(str(uuid()))
        if perm_name:
            user.user_permissions.add(pm.get_perm_model(perm_name))
            user.save()
        return user



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .base import WeChatTestCase


class AdminSiteTestCase(WeChatTestCase):
    def test_admin_view(self):
        """测试admin view"""
        # 测试request能正确拿到appid
        pass

        # 测试响应有extra_context
        pass




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth.models import ContentType, Permission

from ..models import permission as p, WeChatApp
from .base import WeChatTestCase


class PermissionTestCase(WeChatTestCase):
    def test_migrate(self):
        """测试迁移权限"""
        name = "debug"
        p.permissions[name] = "debug %(appname)s permission"
        codename = p.get_perm_name(self.app, name)
        try:
            # 测试upgrade
            p.upgrade_perms((name,))

            content_type = ContentType.objects.get_for_model(WeChatApp)
            Permission.objects.get(
                content_type=content_type,
                codename=codename
            )

            # 测试downgrade
            p.downgrade_perms((name,))
            count = Permission.objects.filter(
                content_type=content_type,
                codename=codename
            ).count()
            self.assertEqual(count, 0)
            count = Permission.objects.filter(
                content_type=content_type,
            ).count()
            self.assertNotEqual(count, 0)
        finally:
            p.permissions.pop(name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from django.utils.module_loading import import_string
from wechatpy import exceptions as excs, WeChatClient as _Client
from wechatpy.constants import WeChatErrorCode
from wechatpy.client import api

from . import settings


class WeChatMaterial(api.WeChatMaterial):
    def get_raw(self, media_id):
        return self._post(
            'material/get_material',
            data={
                'media_id': media_id
            }
        )


class WeChatMessage(api.WeChatMessage):
    def send_articles(self, user_id, articles, account=None):
        try:
            return super(WeChatMessage, self).send_articles(
                user_id, articles, account)
        except excs.WeChatClientException as e:
            # 对于主动发送图文消息,只发送一条
            # 参见 https://mp.weixin.qq.com/cgi-bin/announce?action=getannouncement&announce_id=115383153198yAvN
            if e.errcode == WeChatErrorCode.INVALID_MESSAGE_TYPE:
                return super(WeChatMessage, self).send_articles(
                    user_id, articles[:1], account)
            raise


class WeChatClient(_Client):
    """继承原有WeChatClient添加日志功能 追加accesstoken url获取"""
    # 增加raw_get方法
    material = WeChatMaterial()
    message = WeChatMessage()

    ACCESSTOKEN_URL = None

    def __init__(self, app):
        """:type app: wechat_django.models.WeChatApp"""
        session = import_string(settings.SESSIONSTORAGE)
        if callable(session):
            session = session(app)
        self.app = app
        if app.configurations.get("ACCESSTOKEN_URL"):
            self.ACCESSTOKEN_URL = app.configurations["ACCESSTOKEN_URL"]
        super(WeChatClient, self).__init__(
            app.appid, app.appsecret, session=session)

    def _fetch_access_token(self, url, params):
        """自定义accesstoken url"""
        return super(WeChatClient, self)._fetch_access_token(
            self.ACCESSTOKEN_URL or url, params)

    def _request(self, method, url_or_endpoint, **kwargs):
        self._update_log(method=method, url=url_or_endpoint, **kwargs)
        try:
            rv = super(WeChatClient, self)._request(
                method, url_or_endpoint, **kwargs)
            self._log(logging.DEBUG)
            return rv
        except Exception as e:
            if isinstance(e, excs.WeChatClientException):
                self._log(logging.WARNING)
            else:
                self._log(logging.ERROR)
            raise

    def _handle_result(self, res, method=None, url=None, *args, **kwargs):
        resp = res.content if hasattr(res, "content") else res
        self._update_log(resp=resp)
        return super(WeChatClient, self)._handle_result(
            res, method, url, *args, **kwargs)

    def _update_log(self, **kwargs):
        if not hasattr(self, "_log_kwargs"):
            self._log_kwargs = dict()
        self._log_kwargs.update(kwargs)

    def _log(self, level):
        msg = "{method}\t{url}".format(
            method=self._log_kwargs.get("method", ""),
            url=self._log_kwargs.get("url", "")
        )
        for k, v in self._log_kwargs.items():
            msg += "\t{k}: {v}".format(k=k, v=v)
        kwargs = dict()
        if level >= logging.WARNING:
            kwargs["exc_info"] = True
        self.app.logger("api").log(level, msg)
        self._log_kwargs.clear()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4 as uuid

from django.contrib.admin import site
from django.contrib.auth.models import ContentType, Permission, User
from django.urls import ResolverMatch

from ..models import permission as pm, WeChatApp
from ..pay.admin.payapp import WeChatPayInline, WeChatAppWithPayAdmin
from ..pay.models.app import WeChatPay
from .base import WeChatTestCase


class PermissionTestCase(WeChatTestCase):
    def test_permission_grant(self):
        """测试授权"""
        perm_names = pm.list_perm_names(self.app)
        self.assertEqual(len(perm_names), len(pm.permissions) + 1)

        # 测试相关权限是否一起授权
        for perm_name in perm_names:
            user = self._create_user(perm_name)
            needed_perms = pm.get_require_perm_names(self.app.name, perm_name)
            for permission in needed_perms:
                self.assertHasPermission(user, permission)

        # 测试组授权
        pass

    def test_index_menu(self):
        """测试首页菜单权限"""
        def assertMenuCorrect(perm_name, manage=False, apps=None):
            request = self.rf().get("/admin/")
            request.user = self._create_user(perm_name)

            app_dict = site._build_app_dict(request)

            if apps is None:
                apps = set([self.app.name])

            if manage:
                self.assertEqual(len(app_dict["wechat_django"]["models"]), 1)
            else:
                self.assertNotIn("wechat_django", app_dict)

            if apps:
                appnames = {
                    app["object_name"]
                    for app in app_dict["wechat_django_apps"]["models"]
                }
                self.assertEqual(apps, appnames)
            else:
                self.assertNotIn("wechat_django_apps", app_dict)

        perm_names = pm.list_perm_names(self.app)

        # 拥有全部权限
        all_perm_name = pm.get_perm_name(self.app)
        assertMenuCorrect(all_perm_name, manage=True)

        # 仅拥有管理权限
        manage_perm_name = pm.get_perm_name(self.app, "manage")
        manage_permission = manage_perm_name
        assertMenuCorrect(manage_permission, manage=True, apps=[])

        for perm_name in perm_names - set([manage_perm_name, all_perm_name]):
            assertMenuCorrect(perm_name, manage=False)

    def test_app_menu(self):
        """测试各微信号菜单权限"""
        def assertMenuCorrect(perm_name):
            request = self.rf().get("/admin/wechat_django/apps/" + str(self.app.id))
            request.user = self._create_user(perm_name)
            request.app = self.app
            request.app_id = self.app.id

            app_dict = site._build_app_dict(request, "wechat_django")
            permissions = pm.get_user_permissions(
                request.user, self.app, exclude_manage=True, exclude_sub=True)
            # 只有子权限是不会拥有菜单权限的 遂去掉子权限
            if permissions:
                all_permissions = {
                    model["object_name"].lower().replace("wechat", "")
                    for model in app_dict["models"]
                    if list(filter(None, model["perms"].values()))
                }
                self.assertEqual(permissions, all_permissions)
            else:
                self.assertIsNone(app_dict)

        for perm_name in pm.list_perm_names(self.app):
            assertMenuCorrect(perm_name)

    def assertHasPermission(self, user, permission):
        permission = pm.get_perm_model(permission)
        self.assertIn(permission, user.user_permissions.all())

    def _create_user(self, perm_name=None):
        user = User.objects.create_user(str(uuid()))
        if perm_name:
            user.user_permissions.add(pm.get_perm_model(perm_name))
            user.save()
        return user



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from wechatpy.exceptions import WeChatException


class BadMessageRequest(ValueError):
    pass


class MessageHandleError(ValueError):
    pass


class WeChatAbilityError(WeChatException):
    """当公众号尝试使用不具备的某一能力时抛出的异常"""

    API = (10000, _("没有API访问能力"))
    TEMPLATE = (10001, _("没有模板消息能力"))
    OAUTH = (10002, _("没有OAUTH能力"))
    INTERACTABLE = (10003, _("没有消息交互能力"))

    def __init__(self, err):
        super(WeChatAbilityError, self).__init__(*err)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from copy import deepcopy

from django.db import models as m
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
import requests
from six import text_type
from wechatpy import replies

from ..exceptions import MessageHandleError
from ..utils.model import enum2choices, model_fields
from . import (
    Article, Material, MessageHandler, MsgType as BaseMsgType, WeChatModel)


class Reply(WeChatModel):
    class MsgType(BaseMsgType):
        MUSIC = "music"
        NEWS = "news"

        CUSTOM = "custom" #  自定义业务
        FORWARD = "forward"  #  转发

    handler = m.ForeignKey(
        MessageHandler, related_name="replies", on_delete=m.CASCADE)

    type = m.CharField(
        _("type"), db_column="type", max_length=16,
        choices=enum2choices(MsgType))
    _content = JSONField(db_column="content", default=dict)

    weight = m.IntegerField(_("weight"), default=0, null=False)
    created_at = m.DateTimeField(_("created_at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated_at"), auto_now=True)

    class Meta(object):
        verbose_name = _("reply")
        verbose_name_plural = _("replies")

        ordering = ("-weight", "id")

    @property
    def app(self):
        return self.handler.app

    @property
    def content(self):
        return self._content

    def __init__(self, *args, **kwargs):
        field_names = model_fields(self)
        content_keys = set(kwargs.keys()) - field_names
        content = dict()
        for key in content_keys:
            content[key] = kwargs.pop(key)
        kwargs["_content"] = content
        super(Reply, self).__init__(*args, **kwargs)

    def send(self, message_info):
        """主动回复
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        reply = self.reply(message_info)
        funcname, kwargs = self.reply2send(reply)
        func = funcname and getattr(self.app.client.message, funcname)
        return func and func(**kwargs)

    def reply(self, message_info):
        """被动回复
        :type message_info: wechat_django.models.WeChatMessageInfo
        :rtype: wechatpy.replies.BaseReply
        """
        if self.type == self.MsgType.FORWARD:
            # 转发业务
            reply = self.reply_forward(message_info)
        elif self.type == self.MsgType.CUSTOM:
            # 自定义业务
            reply = self.reply_custom(message_info)
        else:
            # 正常回复类型
            reply = self.normal_reply(message_info.message)
        return reply

    def reply_forward(self, message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        resp = requests.post(
            self.content["url"], message_info.raw,
            params=message_info.request.GET, timeout=4.5)
        resp.raise_for_status()
        return replies.deserialize_reply(resp.content)

    def reply_custom(self, message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        try:
            func = import_string(self.content["program"])
        except:
            raise MessageHandleError("custom bussiness not found")
        else:
            appname = message_info.app.name
            message = message_info.message
            if not hasattr(func, "message_handler"):
                e = "handler must be decorated by wechat_django.handler.message_handler"
                raise MessageHandleError(e)
            elif (hasattr(func.message_handler, "__contains__")
                and appname not in func.message_handler):
                e = "this handler cannot assigned to {0}".format(appname)
                raise MessageHandleError(e)
            reply = func(message_info)
            if not reply:
                return ""
            elif isinstance(reply, text_type):
                reply = replies.TextReply(content=reply)
            reply.source = message.target
            reply.target = message.source
            return reply

    def normal_reply(self, message):
        """
        :type message: wechatpy.messages.BaseMessage
        """
        if self.type == self.MsgType.NEWS:
            klass = replies.ArticlesReply
            media = self.app.materials.get(media_id=self.content["media_id"])
            # 将media_id转为content
            data = dict(
                articles=media.articles_json,
                media_id=self.content["media_id"]
            )
        elif self.type == self.MsgType.MUSIC:
            klass = replies.MusicReply
            data = dict(**self.content)
        elif self.type == self.MsgType.VIDEO:
            klass = replies.VideoReply
            data = dict(**self.content)
        elif self.type == self.MsgType.IMAGE:
            klass = replies.ImageReply
            data = dict(media_id=self.content["media_id"])
        elif self.type == self.MsgType.VOICE:
            klass = replies.VoiceReply
            data = dict(media_id=self.content["media_id"])
        else:
            klass = replies.TextReply
            data = dict(content=self.content["content"])
        return klass(message=message, **data)

    @staticmethod
    def reply2send(reply):
        """
        将主动回复生成的reply转换为被动回复的方法及变量
        :type reply: wechatpy.replies.BaseReply
        """
        if not reply or isinstance(reply, replies.EmptyReply):
            return None, None
        type = ""
        kwargs = dict(user_id=reply.target)
        if isinstance(reply, replies.ArticlesReply):
            kwargs["articles"] = deepcopy(reply.articles)
            for article in kwargs["articles"]:
                article["picurl"] = article["image"]
                del article["image"]
            type = "articles"
        elif isinstance(reply, replies.MusicReply):
            kwargs["url"] = reply.music_url
            kwargs["hq_url"] = reply.hq_music_url
            kwargs["thumb_media_id"] = reply.thumb_media_id
            kwargs["title"] = reply.title
            kwargs["description"] = reply.description
        elif isinstance(reply, replies.VideoReply):
            kwargs["media_id"] = reply.media_id
            kwargs["title"] = reply.title
            kwargs["description"] = reply.description
        elif isinstance(reply, (replies.ImageReply, replies.VoiceReply)):
            kwargs["media_id"] = reply.media_id
        elif isinstance(reply, replies.TextReply):
            kwargs["content"] = reply.content
        else:
            raise MessageHandleError("unknown reply type")
        type = type or reply.type
        funcname = "send_" + type
        return funcname, kwargs

    @classmethod
    def from_mp(cls, app, data):
        type = data["type"]
        if type == "img":
            type = cls.MsgType.IMAGE
        elif type == cls.MsgType.VIDEO:
            # video是链接
            type = cls.MsgType.TEXT
            data = dict(content='<a href="{0}">{1}</a>'.format(
                data["content"], _("video")))

        kwargs = dict(type=type)
        if type == cls.MsgType.TEXT:
            kwargs.update(content=data["content"])
        elif type in (cls.MsgType.IMAGE, cls.MsgType.VOICE):
            # 按照文档 是临时素材 需要转换为永久素材
            media_id = Material.as_permenant(data["content"], app, False)
            kwargs.update(media_id=media_id)
        elif type == cls.MsgType.NEWS:
            media_id = data["content"]
            # 同步图文
            Article.sync(app, media_id)
            kwargs.update(
                media_id=media_id,
                content=data["news_info"]["list"]
            )
        else:
            raise ValueError("unknown reply type {0}".format(type))
        return cls(**kwargs)

    @classmethod
    def from_menu(cls, menu, data):
        app = menu.app
        type = data["type"]
        if type == "img":
            type = cls.MsgType.IMAGE
        elif type == cls.MsgType.VIDEO:
            # video是链接
            type = cls.MsgType.TEXT
            data = dict(content='<a href="{0}">{1}</a>'.format(
                data["value"], data.get("name", _("video"))))

        kwargs = dict(type=type)
        if type == cls.MsgType.TEXT:
            kwargs.update(content=data["content"])
        elif type in (cls.MsgType.IMAGE, cls.MsgType.VOICE):
            media_id = Material.as_permenant(data["value"], app, False) if app\
                else data["value"]
            kwargs.update(media_id=media_id)
        elif type == cls.MsgType.NEWS:
            media_id = data["value"]
            # 同步图文
            app.sync_articles(media_id)
            kwargs.update(
                media_id=media_id,
                content=data["news_info"]["list"]
            )
        else:
            raise ValueError("unknown menu reply type {0}".format(type))
        return cls(**kwargs)

    def __str__(self):
        if self.handler_id:
            return "{0} - {1}".format(self.handler.name, self.type)
        return "{0}".format(self.type)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m, transaction
from django.utils.translation import ugettext_lazy as _

from ..exceptions import WeChatAbilityError
from ..utils.model import model_fields
from . import appmethod, WeChatApp, WeChatModel


class Template(WeChatModel):
    app = m.ForeignKey(
        WeChatApp, related_name="templates", on_delete=m.CASCADE)

    template_id = m.CharField(_("template_id"), max_length=64)
    alias = m.CharField(
        _("alias"), max_length=16, blank=True, null=True,
        help_text=_("模板别名,用于程序快速查询模板,单app下唯一"))

    title = m.CharField(_("title"), max_length=32)
    content = m.TextField(_("content"))
    example = m.TextField(_("example"), blank=True, null=True)

    primary_industry = m.CharField(
        _("primary industry"), max_length=64, blank=True, null=True)
    deputy_industry = m.CharField(
        _("deputy industry"), max_length=64, blank=True, null=True)

    comment = m.TextField(_("remark"), blank=True, null=True)
    created_at = m.DateTimeField(_("created at"), auto_now_add=True)

    class Meta:
        verbose_name = _("template")
        verbose_name_plural = _("templates")

        unique_together = (("app", "template_id"), ("app", "alias"))

    @classmethod
    @appmethod("sync_templates")
    def sync(cls, app):
        """
        同步微信模板
        :type app: wechat_django.models.WeChatApp
        """
        if app.type == WeChatApp.Type.SERVICEAPP:
            resp = app.client.template.get_all_private_template()
            templates = resp["template_list"]
        elif app.type == WeChatApp.Type.MINIPROGRAM:
            templates = list(cls._iter_wxa_templates(app))
        else:
            raise WeChatAbilityError(WeChatAbilityError.TEMPLATE)

        with transaction.atomic():
            (app.templates
                .exclude(template_id__in=[t["template_id"] for t in templates])
                .delete())
            rv = []
            for t in templates:
                defaults = dict(app=app)
                defaults.update({
                    k: v for k, v in t.items() if k in model_fields(cls)})
                template = app.templates.update_or_create(
                    app=app, template_id=t["template_id"],
                    defaults=defaults
                )[0]
                rv.append(template)
            return rv

    def send(
        self, user, data=None, url=None, appid=None, pagepath=None, page=None,
        form_id=None, emphasis_keyword=None, **kwargs):
        """
        发送模板消息
        :param user: 用户openid或用户对象
        :type user: wechat_django.models.WeChatUser or str
        :param data: 接口的data字段
        :param url: 跳转的url 仅服务号有效
        :param appid: 跳转小程序的appid 仅服务号有效
        :param pagepath: 跳转小程序的具体页
        :param page: 跳转小程序的具体页 pagepath的别名
        :param form_id: 发送模板消息需要的form_id 仅小程序有效
        :param emphasis_keyword: 模板需要放大的关键词，不填则默认无放大
        :param kwargs: 接口的data字段 可直接传字符串 当data未填写时 使用该字段
        """
        from . import WeChatUser

        openid = user.openid if isinstance(user, WeChatUser) else user
        pagepath = pagepath or page
        if not data:
            data = {
                k: v if isinstance(v, dict) else dict(value=v)
                for k, v in kwargs.items()
            }

        if self.app.type == WeChatApp.Type.SERVICEAPP:
            return self._send_service(openid, data, url, appid, pagepath)
        elif self.app.type == WeChatApp.Type.MINIPROGRAM:
            return self._send_miniprogram(
                openid, data, form_id, pagepath, emphasis_keyword)
        else:
            raise WeChatAbilityError(WeChatAbilityError.TEMPLATE)

    def _send_service(
        self, openid, data, url=None, appid=None, pagepath=None):
        """发送服务号模板消息"""
        miniprogram = dict(
            appid=appid,
            pagepath=pagepath
        ) if appid else None
        return self.app.client.message.send_template(
            openid, self.template_id, data, url=url, mini_program=miniprogram)

    def _send_miniprogram(
        self, openid, data, form_id, page=None, emphasis_keyword=None):
        """发送小程序模板消息"""
        return self.app.client.send_template_message(
            openid, self.template_id, data, form_id, page=page,
            emphasis_keyword=emphasis_keyword)

    @staticmethod
    def _iter_wxa_templates(app):
        while True:
            offset = 0
            count = 20
            templates = app.client.list_templates(offset, count)
            if not templates:
                return
            for template in templates:
                yield template
            if len(templates) < count:
                # 未满说明没有下一页了
                return
            offset += count

    def save(self, *args, **kwargs):
        self.alias = self.alias or None
        return super(Template, self).save(*args, **kwargs)

    def __str__(self):
        return "{title} ({template_id})".format(
            title=self.title, template_id=self.template_id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re

from django.db import models as m
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField

from ..utils.model import enum2choices, model_fields
from . import MessageHandler, MsgType, WeChatModel


class Rule(WeChatModel):
    class Type(object):
        MSGTYPE = "msg_type"  # 类型匹配
        EVENT = "event"  # 事件
        EVENTKEY = "eventkey"  # 指定事件
        CONTAIN = "contain"  # 包含
        EQUAL = "equal"  # 匹配
        REGEX = "regex"  # 正则
        ALL = "all"  # 全部

    class ReceiveMsgType(MsgType):
        LOCATION = "location"
        LINK = "link"
        SHORTVIDEO = "shortvideo"
        EVENT = "event"

    handler = m.ForeignKey(
        MessageHandler, related_name="rules", null=False, on_delete=m.CASCADE)

    type = m.CharField(
        _("type"), max_length=16, choices=enum2choices(Type))  # 规则类型
    _content = JSONField(db_column="content", blank=True)  # 规则内容

    weight = m.IntegerField(_("weight"), default=0, null=False)
    created_at = m.DateTimeField(_("created_at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated_at"), auto_now=True)

    @property
    def content(self):
        return self._content

    class Meta:
        verbose_name = _("rule")
        verbose_name_plural = _("rules")

        ordering = ("-weight", "id")

    def __init__(self, *args, **kwargs):
        content_keys = set(kwargs.keys()) - model_fields(self)
        content = dict()
        for key in content_keys:
            content[key] = kwargs.pop(key)
        kwargs["_content"] = content
        super(Rule, self).__init__(*args, **kwargs)

    def match(self, message_info):
        """
        :type message_info: wechat_django.models.WeChatMessageInfo
        """
        return self._match(message_info.message)

    def _match(self, message):
        """
        :type message: wechatpy.messages.BaseMessage
        """
        if self.type == self.Type.ALL:
            return True
        elif self.type == self.Type.MSGTYPE:
            return message.type == self.content["msg_type"]
        elif self.type == self.Type.EVENT:
            return (message.type == self.ReceiveMsgType.EVENT
                and self._event_match(message))
        elif self.type == self.Type.EVENTKEY:
            return (message.type == self.ReceiveMsgType.EVENT
                and self._event_match(message)
                and hasattr(message, "key")
                and message.key == self.content["key"])
        elif self.type == self.Type.CONTAIN:
            return (message.type == self.ReceiveMsgType.TEXT
                and self.content["pattern"] in message.content)
        elif self.type == self.Type.EQUAL:
            return (message.type == self.ReceiveMsgType.TEXT
                and message.content == self.content["pattern"])
        elif self.type == self.Type.REGEX:
            return (message.type == self.ReceiveMsgType.TEXT
                and re.search(self.content["pattern"], message.content))
        return False

    def _event_match(self, message):
        event = message.event.lower()
        target = self.content["event"].lower()
        if target == MessageHandler.EventType.SUBSCRIBE:
            # wechatpy对eventtype进行了二次封装
            return event in (
                MessageHandler.EventType.SUBSCRIBE, "subscribe_scan")
        else:
            return event == target

    @classmethod
    def from_mp(cls, data, handler=None):
        return cls(
            handler=handler,
            type=data["match_mode"],
            pattern=data["content"]
        )

    def __str__(self):
        if self.handler_id:
            return "{0} - {1}".format(self.handler.name, self.type)
        return "{0}".format(self.type)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.http import response

from ..sites.wechat import WeChatSite, WeChatViewSet
from .base import WeChatTestCase


class WeChatSiteTestCase(WeChatTestCase):
    def test_patch_request(self):
        """测试patch_request,request能拿到wechat对象,并且各属性均正确"""
        # 测试基本的WeChatHttpRequest
        pass

        # 测试WeChatMessageRequest
        pass

        # 测试WeChatOAuthRequest
        pass

    def test_app_queryset(self):
        """测试app_queryset正确"""
        that = self
        class TestSite(WeChatSite):
            @property
            def app_queryset(self):
                return super(TestSite, self).app_queryset.filter(
                    name=that.app.name)

        class TestViewSet(WeChatViewSet):
            def test_view(self, request):
                return response.HttpResponse(status=204)
        
        # 在app_queryset中的公众号可访问,否则404
        site = TestSite()
        site.register(TestViewSet)
        viewset = site.get_registered(TestViewSet)
        view = viewset.wechat_view(viewset.test_view)
        resp = view(self.rf().get("/"), self.app.name)
        self.assertEqual(resp.status_code, 204)
        self.assertRaises(
            response.Http404, view, self.rf().get("/"), self.another_app.name)

    def test_wechat_view(self):
        """测试wechat_view"""
        that = self
        class ViewSet(WeChatViewSet):
            def test_view(self, request):
                return response.HttpResponse(status=204)

            def test_request_view(self, request):
                that.assertEqual(request.wechat.app.id, that.app.id)
                that.assertEqual(request.wechat.appname, that.app.name)
                return response.HttpResponse(status=204)

        # 测试http method正确
        viewset = ViewSet(WeChatSite())
        view = viewset.wechat_view(viewset.test_view, methods=("POST",))
        resp = view(self.rf().get("/"), self.app.name)
        self.assertEqual(resp.status_code, 405)
        resp = view(self.rf().post("/"), self.app.name)
        self.assertEqual(resp.status_code, 204)

        # 测试原view经过wechat_view后request能拿到WeChatInfo对象
        view = viewset.wechat_view(viewset.test_request_view)
        resp = view(self.rf().get("/"), self.app.name)
        self.assertEqual(resp.status_code, 204)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .base import WeChatTestCase


class SettingTestCase(WeChatTestCase):
    def test_settings(self):
        """测试配置是否有效"""
        pass
